import os.path

import numpy as np

import util.plot


def distribution_time(measurement_dict, file='/tmp/distribution_time.png', time_step=1/1., t_min=None, t_max=None, tick_power=None):
    m = measurement_dict
    m.discard_space()
    m.categorize_indices((time_step,))

    n = m.numbers()
    t = n[:,0]
    if t_min is None:
        t_min = t.min()
    if t_max is None:
        t_max = t.max()
    n = n[:,4]

    util.plot.histogram(t, file, weights=n, bins=range(int(t_min),int(t_max+2)), tick_power=tick_power)


def distribution_year(measurement_dict, file='/tmp/distribution_year.png', time_step=1/365., tick_power=None):
    m = measurement_dict
    m.discard_space()
    m.discard_year()
    m.categorize_indices((time_step,))
    n = m.numbers()
    t = n[:,0] / time_step
    n = n[:,4]
    util.plot.histogram(t, file, weights=n, bins=range(0, int(np.ceil(1/time_step))+1), tick_power=tick_power)


def distribution_depth(measurement_dict, file='/tmp/distribution_depth.png', z_max=None, bin_size=50, use_log_scale=True):
    m = measurement_dict
    m.dicard_keys((0,1,2))
    n = m.numbers()
    z = n[:,3]
    if z_max is None:
        z_max = z.max()
    n = n[:,4]
    util.plot.histogram(z, file, weights=n, bins=np.arange(0, z_max+1, bin_size), use_log_scale=use_log_scale)


def value_histograms(measurement_dict, lsm, file='/tmp/value_histogram_{}.png', min_measurements=1, step_size=0.01):
    measurement_dict.discard_year()
    measurement_dict.coordinates_to_map_indices(lsm, int_indices=True)
    n = measurement_dict.numbers(min_measurements=min_measurements)
    for i in range(len(n)):
        measurement_index = n[i,:-1]
        util.plot.histogram(measurement_dict[measurement_index], file.format(measurement_index), step_size=step_size)


def _prepare_filename(file, lsm=None, insertion=None):
    file_prefix, file_ext  = os.path.splitext(file)
    if lsm is not None:
        file_prefix = file_prefix + '_' + str(lsm)
    if insertion is not None:
        file_prefix = file_prefix + '_' + insertion

    return file_prefix + file_ext


def _plot_map(data, lsm, file, layer=None, v_min=None, v_max=None, use_log_scale=False, colorbar_kwargs={'fraction':0.021, 'pad':0.05, 'aspect':20, 'orientation':'vertical'}):
    data = lsm.insert_index_values_in_map(data, no_data_value=np.inf)
    if layer is not None:
        data = data[:, :, :, layer]
        data = data.reshape(data.shape + (1,))
    file = _prepare_filename(file, lsm)
    util.plot.data(data, file, no_data_value=np.inf, v_min=v_min, v_max=v_max, use_log_scale=use_log_scale, contours=False, colorbar=True, power_limit=0, colorbar_kwargs=colorbar_kwargs)

def _plot_histogram(data, lsm, file, bins=None, step_size=None, v_min=None, v_max=None, use_log_scale=False, tick_power=None):
    file = _prepare_filename(file, lsm, 'histogram')
    util.plot.histogram(data, file, bins=bins, step_size=step_size, x_min=v_min, x_max=v_max, use_log_scale=use_log_scale, tick_power=tick_power)

def _plot(data, lsm, file, bins=None, step_size=None, layer=None, v_min=None, v_max=None, use_log_scale=False, type='both'):
    if type == 'both' or type == 'map':
        _plot_map(data, lsm, file, layer=layer, v_min=v_min, v_max=v_max, use_log_scale=use_log_scale)
    if type == 'both' or type == 'histogram':
        _plot_histogram(data[:,-1], lsm, file, bins=bins, step_size=step_size, v_min=v_min, v_max=v_max, use_log_scale=use_log_scale)


def distribution_space(measurement_dict, lsm, file='/tmp/distribution_space.png', layer=None, use_log_scale=True, type='both'):
    ## prepare data
    measurement_dict.discard_year()
    measurement_dict.coordinates_to_map_indices(lsm, int_indices=True)
    data = measurement_dict.numbers()
    ## set v_min and v_max
    if use_log_scale:
        v_min = 1
        v_max = 10**np.min([np.floor(np.log(np.nanmax(data))), 3])
    else:
        v_min = 1
        v_max = None
    ## colorbar ticks
    colorbar_kwargs = {'fraction':0.021, 'pad':0.05, 'aspect':20, 'orientation':'vertical'}
    data_max = data[:,-1].max()
    if data_max < 10:
        colorbar_kwargs['ticks'] = np.arange(1, data_max+1)
    ## plot
    _plot_map(data, lsm, file, layer=layer, v_min=v_min, v_max=v_max, use_log_scale=use_log_scale, colorbar_kwargs=colorbar_kwargs)
    _plot_histogram(data[:,-1], lsm, file, step_size=1, v_min=1, v_max=200, use_log_scale=use_log_scale)


def sample_mean(measurement_dict, lsm, file='/tmp/sample_mean.png', v_max=None, layer=None, type='both'):
    measurement_dict.discard_year()
    measurement_dict.coordinates_to_map_indices(lsm, int_indices=True)
    data = measurement_dict.means()
    _plot_map(data, lsm, file, layer=layer, v_min=0, v_max=v_max)
    _plot_histogram(data[:,-1], lsm, file, step_size=0.1, v_min=0, v_max=4, tick_power=3)


def sample_deviation(measurement_dict, lsm, file='/tmp/sample_deviation.png', v_max=None, layer=None, type='both', min_measurements=5):
    measurement_dict.discard_year()
    measurement_dict.coordinates_to_map_indices(lsm, int_indices=True)
    data = measurement_dict.deviations(min_values=min_measurements)
    _plot_map(data, lsm, file, layer=layer, v_min=0, v_max=v_max)
    _plot_histogram(data[:,-1], lsm, file, step_size=0.05, v_min=0, v_max=2, use_log_scale=True)


def normalized_values(measurement_dict, lsm, year, file='/tmp/normalized_values.png', layer=None, type='both', min_measurements=5):
    measurement_dict.normalize_with_lsm(lsm, min_values=min_measurements)
    measurement_dict.filter_year(year)
    measurement_dict.discard_year()
    measurement_dict.coordinates_to_map_indices(lsm, int_indices=True)
    data = measurement_dict.means()
    _plot_map(data, lsm, file, layer=layer, v_min=-1, v_max=1)
    _plot_histogram(data[:,-1], lsm, file, step_size=0.05, use_log_scale=True)




import numpy as np
import scipy.sparse

import util.math.util
import util.cache.file
import util.logging

import measurements.universal.dict
import measurements.universal.constants
import measurements.constants

logger = util.logging.logger



class SampleMeanAndDeviation():
    
    def __init__(self, points, values, sample_lsm):
        self.points = points
        self.values = values
        self.sample_lsm = sample_lsm

    @property
    def measurement_dict(self):
        measurement_dict = measurements.universal.dict.MeasurementsDict()
        measurement_dict.append_values(self.points, self.values)
        return measurement_dict

    
    ## general
    
    def _convert_map_indices_dict_to_array_for_points(self, data_dict, is_discard_year):
        points = self.sample_lsm.coordinates_to_map_indices(self.points, discard_year=is_discard_year, int_indices=True)
        n = len(points)
        point_data = np.ma.masked_all(n)
        for i in range(n):
            try:
                value_list_i = data_dict[points[i]]
            except KeyError:
                pass
            else:
                assert len(value_list_i) == 1
                point_data[i] = value_list_i[0]
        return point_data


    def _convert_map_indices_dict_to_points_and_values_array(self, data_dict):
        data_dict.map_indices_to_coordinates(self.sample_lsm)
        return data_dict.toarray()


    def _sample_data_dict_concentration_based(self, data_function):
        data_dict = self.measurement_dict
        data_dict.coordinates_to_map_indices(self.sample_lsm, int_indices=True)
        data_dict.means(min_number_of_values=1, return_type='self')
        data_dict.discard_year()
        data_function(data_dict)
        return data_dict
    
    def _sample_data_dict_average_noise_based(self, data_function, return_values_at_points=True):
        data_dict = self.measurement_dict
        data_dict.coordinates_to_map_indices(self.sample_lsm, int_indices=True)
        data_function(data_dict)
        data_dict.discard_year()
        data_dict.means(min_number_of_values=1, return_type='self')
        return data_dict


    def _sample_data_dict_noise_based(self, data_function, return_values_at_points=True):
        data_dict = self.measurement_dict
        data_dict.coordinates_to_map_indices(self.sample_lsm, int_indices=True)
        data_function(data_dict)
        return data_dict

    
    ## mean

    def sample_concentration_means_map_indices_dict(self, min_measurements=measurements.constants.MEAN_MIN_MEASUREMENTS):
        logger.debug('Calculating sample_concentration_means_map_indices_dict with min_measurements {}.'.format(min_measurements))
        data_function = lambda data_dict: data_dict.means(min_number_of_values=min_measurements, return_type='self')
        data = self._sample_data_dict_concentration_based(data_function)
        return data

    def sample_concentration_means(self, min_measurements=measurements.constants.MEAN_MIN_MEASUREMENTS):
        logger.debug('Calculating sample_concentration_means with min_measurements {}.'.format(min_measurements))
        data_dict = self.sample_concentration_means_map_indices_dict(min_measurements=min_measurements)
        data = self._convert_map_indices_dict_to_array_for_points(data_dict, is_discard_year=True)
        return data
    

    ## deviation
    
    def sample_concentration_standard_deviations_map_indices_dict(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_concentration_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_function = lambda data_dict: data_dict.standard_deviations(min_number_of_values=min_measurements, min_value=min_value, return_type='self')
        data_dict = self._sample_data_dict_concentration_based(data_function) 
        return data_dict

    
    def sample_concentration_standard_deviations(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_concentration_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_dict = self.sample_concentration_standard_deviations_map_indices_dict(min_measurements=min_measurements, min_value=min_value)
        data = self._convert_map_indices_dict_to_array_for_points(data_dict, is_discard_year=True)
        return data


    def sample_average_noise_standard_deviations_map_indices_dict(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_average_noise_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_function = lambda data_dict: data_dict.standard_deviations(min_number_of_values=min_measurements, min_value=min_value, return_type='self')
        data_dict = self._sample_data_dict_average_noise_based(data_function)
        return data_dict


    def sample_average_noise_standard_deviations(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_average_noise_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_dict = self.sample_average_noise_standard_deviations_map_indices_dict(min_measurements=min_measurements, min_value=min_value)
        data = self._convert_map_indices_dict_to_array_for_points(data_dict, is_discard_year=True)
        return data

    
    def sample_noise_standard_deviations_map_indices_dict(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_noise_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_function = lambda data_dict: data_dict.standard_deviations(min_number_of_values=min_measurements, min_value=min_value, return_type='self')
        data_dict = self._sample_data_dict_noise_based(data_function)
        return data_dict

    
    def sample_noise_standard_deviations(self, min_measurements=measurements.constants.DEVIATION_MIN_MEASUREMENTS, min_value=0):
        logger.debug('Calculating sample_noise_standard_deviations with min_measurements {} and min_value {}.'.format(min_measurements, min_value))
        data_dict = self.sample_noise_standard_deviations_map_indices_dict(min_measurements=min_measurements, min_value=min_value)
        data = self._convert_map_indices_dict_to_array_for_points(data_dict, is_discard_year=False)
        return data
    
    

class SampleCorrelationMatrix:
    
    def __init__(self, measurements, sample_lsm, min_measurements, max_year_diff=float('inf'), min_abs_correlation=measurements.universal.constants.CORRELATION_MIN_ABS_VALUE, max_abs_correlation=measurements.universal.constants.CORRELATION_MAX_ABS_VALUE, dtype=np.float32, matrix_format='csc'):
        
        self.measurements = measurements        
        self.sample_lsm = sample_lsm

        self.min_measurements = min_measurements
        self.max_year_diff = max_year_diff
        self.min_abs_correlation = min_abs_correlation
        self.max_abs_correlation = max_abs_correlation

        self.dtype = np.dtype(dtype)
        self.matrix_format = matrix_format


    ## properties

    @property
    def shape(self):
        n = len(self.measurements.points)
        return (n, n)


    ## standard deviation

    @property
    def standard_deviations(self):
        standard_deviations = self.measurements.standard_deviations
        assert np.all(standard_deviations > 0)
        return standard_deviations
    

    ##  map index to point index dict

    def map_indices_to_point_index_dict(self, discard_year=False):
        logger.debug('Calculating map index to point index dict with discard year {}.'.format(discard_year))

        points = self.measurements.points
        map_indices_to_point_index_dict = measurements.universal.dict.MeasurementsDict()

        for i in range(len(points)):
            point = points[i]
            map_index = self.sample_lsm.coordinates_to_map_indices(point, discard_year=discard_year, int_indices=True)
            map_indices_to_point_index_dict.append_value(map_index, i)

        return map_indices_to_point_index_dict
    

    ## same box correlation
    
    @property
    def same_box_correlation_matrix_lower_triangle(self):
        logger.debug('Calculating same box correlation matrix lower triangle with minimal absolute correlation {} in matrix format {} with dtype {}.'.format(self.min_abs_correlation, self.matrix_format, self.dtype))
        
        ## create lil matrix
        correlation_matrix = scipy.sparse.lil_matrix(self.shape, dtype=self.dtype)
        
        ## get values
        standard_deviations = self.standard_deviations
        map_indices_to_point_index_dict = self.map_indices_to_point_index_dict(discard_year=False)
        same_box_sample_covariance_index_array = self.measurements.concentration_standard_deviations**2
    
        ## iterate all point pairs in same box
        for key, point_index_list in map_indices_to_point_index_dict.iterator_keys_and_value_lists():
            n = len(point_index_list)
            
            for i in range(n):
                point_index_i = point_index_list[i]
                
                covariance = same_box_sample_covariance_index_array[point_index_i]
                
                for j in range(i+1, n):
                    point_index_j = point_index_list[j]
                    
                    ## calculate correlation
                    assert same_box_sample_covariance_index_array[point_index_i] == same_box_sample_covariance_index_array[point_index_j]
                    correlation = covariance / (standard_deviations[point_index_i] * standard_deviations[point_index_j])
                    assert correlation >= 0 and correlation <= 1

                    ## if abs correlation geq min correlation, insert 
                    if np.abs(correlation) >= self.min_abs_correlation:

                        point_index_min, point_index_max = (min(point_index_i, point_index_j), max(point_index_i, point_index_j))
                        assert point_index_min < point_index_max
                        correlation_matrix[point_index_max, point_index_min] = correlation
                        logger.debug('Same box correlation {} added to same box lower triangle matrix at ({}, {}).'.format(correlation, point_index_max, point_index_min))

            
        ## convert to needed format
        if self.matrix_format != 'lil':
            logger.debug('Converting matrix to format {} and dtype {}.'.format(self.matrix_format, self.dtype))
            correlation_matrix = correlation_matrix.asformat(self.matrix_format).astype(self.dtype)

        logger.debug('Calculated same box correlation lower triangle matrix with {} entries for minimal absolute correlation {} in matrix format {} with dtype {}.'.format(correlation_matrix.nnz, self.min_abs_correlation, self.matrix_format, self.dtype))
        return correlation_matrix

    
    ## different_boxes_sample_covariances
    
    @property
    def concentrations_same_points_except_year_dict(self):
        m = self.measurements.measurements_dict
        m.categorize_indices_to_lsm(self.sample_lsm, discard_year=False)
        m.means(return_type='self')
        return m.filter_same_points_except_year(min_number_of_values=self.min_measurements)

    
    @property
    def sample_covariance_dict(self):
        ms = self.concentrations_same_points_except_year_dict
        covariance = ms.correlation_or_covariance('covariance', min_number_of_values=self.min_measurements, stationary=False, max_year_diff=self.max_year_diff)
        return covariance
    
    
    def different_boxes_sample_covariances_map_indices_iterator(self):
        different_boxes_sample_covariances_dict = self.sample_covariance_dict
        different_boxes_sample_covariances_dict.coordinates_to_map_indices(self.sample_lsm, int_indices=True)
        for map_indices, value_list in different_boxes_sample_covariances_dict.iterator_keys_and_value_lists():
            assert len(map_indices) == 2
            assert len(value_list) == 1
            values = value_list[0]
            assert len(values) == 2
            quantity, covariance = values
            yield map_indices, quantity, covariance
    
    
    def different_boxes_sample_covariances_point_indices_iterator(self):
        ## get values
        standard_deviations = self.standard_deviations
        points = self.measurements.points
        map_indices_to_point_index_dict = self.map_indices_to_point_index_dict(discard_year=False)
        map_indices_to_point_index_year_discarded_dict = self.map_indices_to_point_index_dict(discard_year=True)
    
        ## iterate over sample covariances
        for map_indices, quantity, covariance in self.different_boxes_sample_covariances_map_indices_iterator():
            assert quantity >= self.min_measurements
            map_indices_array = np.array(map_indices)
            map_indices_diff = map_indices_array[1] - map_indices_array[0]
            logger.debug('Different box sample covariance found with map indices {}, covariance {} and quantity {}.'.format(map_indices, covariance, quantity))
            
            ## iterate over all index pairs with sample covariance
            for point_index_i in map_indices_to_point_index_year_discarded_dict[map_indices[0]]:
                point_i = points[point_index_i]
                point_i_map_index = self.sample_lsm.coordinate_to_map_index(*point_i, discard_year=False, int_indices=True)
                point_j_map_index = tuple(point_i_map_index + map_indices_diff)
                
                try:
                    point_indices_j = map_indices_to_point_index_dict[point_j_map_index]
                except KeyError:
                    pass
                else:
                    for point_index_j in point_indices_j:
                        ## calculate correlation
                        correlation = covariance / (standard_deviations[point_index_i] * standard_deviations[point_index_j])
                        
                        ## return if correlation is big enough
                        if np.abs(correlation) >= self.min_abs_correlation:
                            point_index_min, point_index_max = (min(point_index_i, point_index_j), max(point_index_i, point_index_j))
                            assert point_index_min < point_index_max and point_index_min in (point_index_i, point_index_j) and point_index_max in (point_index_i, point_index_j)

                            yield point_index_min, point_index_max, quantity, correlation
    

    ## different boxes covariance and quantity matrix

    @property
    def different_boxes_quantity_lower_triangle_matrix(self):
        logger.debug('Calculating different boxes quantity lower triangle matrix with minimal absolute correlation {} in matrix format {}.'.format(self.min_abs_correlation, self.matrix_format))

        ## get max quantity
        max_quantity = 0
        for map_indices, quantity, covariance in self.different_boxes_sample_covariances_iterator():
            max_quantity = max(max_quantity, quantity)
        
        ## get matrix dtype
        dtype = util.math.util.min_int_dtype(max_quantity, unsigned=True)
        dtype = np.dtype(dtype)
        
        ## create lil matrix
        quantity_matrix = scipy.sparse.lil_matrix(self.shape, dtype=dtype)
        
        ## insert correlation
        for point_index_min, point_index_max, quantity, correlation in self.different_boxes_sample_covariances_point_indices_iterator():
            assert quantity >= self.min_measurements
            assert np.abs(correlation) >= self.min_abs_correlation
            assert point_index_max > point_index_min
            quantity_matrix[point_index_max, point_index_min] = quantity
        
        ## convert to wanted format
        if self.matrix_format != 'lil':
            logger.debug('Converting quantity matrix to format {}.'.format(self.matrix_format))
            if self.matrix_format == 'csc':
                quantity_matrix = quantity_matrix.asformat('csr')
                logger.debug('quantity matrix converted to format csr.')
            quantity_matrix = quantity_matrix.asformat(self.matrix_format).astype(dtype)
            logger.debug('quantity matrix converted to format {}.'.format(self.matrix_format))
        
        ## return
        logger.debug('Calculated differend boxes quantity lower triangle matrix with {} entries for minimal absolute correlation {}.'.format(correlation_matrix.nnz, self.min_abs_correlation))
        return quantity_matrix
    
    
    @property
    def different_boxes_correlation_lower_triangle_matrix(self):
        logger.debug('Calculating different boxes correlation lower triangle matrix with minimal absolute correlation {} in matrix format {} with dtype {}.'.format(self.min_abs_correlation, self.matrix_format, self.dtype))
        
        ## create lil matrix
        correlation_matrix = scipy.sparse.lil_matrix(self.shape, dtype=self.dtype)
        
        ## insert correlation
        for point_index_min, point_index_max, quantity, correlation in self.different_boxes_sample_covariances_point_indices_iterator():
            assert quantity >= self.min_measurements
            assert np.abs(correlation) >= self.min_abs_correlation
            assert point_index_max > point_index_min
            correlation_matrix[point_index_max, point_index_min] = correlation

        ## convert to wanted format
        if self.matrix_format != 'lil':
            logger.debug('Converting correlation matrix to format {}.'.format(self.matrix_format))
            if self.matrix_format == 'csc':
                correlation_matrix = correlation_matrix.asformat('csr')
                logger.debug('Correlation matrix converted to format csr.')
            correlation_matrix = correlation_matrix.asformat(self.matrix_format).astype(self.dtype)
            logger.debug('Correlation matrix converted to format {}.'.format(self.matrix_format))

        ## return
        logger.debug('Calculated differend boxes correlation lower triangle matrix with {} entries for minimal absolute correlation {} in matrix format {} with dtype {}.'.format(correlation_matrix.nnz, self.min_abs_correlation, self.matrix_format, self.dtype))
        return correlation_matrix


    ## correlation matrix

    @property
    def correlation_matrix(self):
        logger.debug('Calculating correlation matrix for minimal absolute correlation {} and maximal absolute correlation {} in matrix format {} with dtype {}.'.format(self.min_abs_correlation, self.max_abs_correlation, self.matrix_format, self.dtype))
        
        ## add same box and different boxes correlations lower triangle
        correlation_lower_triangle_matrix = self.different_boxes_correlation_lower_triangle_matrix + self.same_box_correlation_matrix_lower_triangle
        assert np.all(np.isclose(correlation_lower_triangle_matrix.diagonal(), 0))  

        ## apply max abs correlation
        mask = np.abs(correlation_lower_triangle_matrix.data) > self.max_abs_correlation
        correlation_lower_triangle_matrix.data[mask] = np.sign(correlation_lower_triangle_matrix.data[mask]) * self.max_abs_correlation
        
        ## add lower and upper triangle
        correlation_matrix = correlation_lower_triangle_matrix + correlation_lower_triangle_matrix.T
        
        ## add diagonal ones
        diagonal = scipy.sparse.eye(correlation_matrix.shape[0])
        correlation_matrix = correlation_matrix + diagonal
        
        ## return
        return correlation_matrix.asformat(self.matrix_format).astype(self.dtype)




class SampleCorrelationMatrixCache(SampleCorrelationMatrix):
    
    ## cachable values

    @util.cache.file.decorator()
    def map_indices_to_point_index_dict(self, discard_year=False):
        return super().map_indices_to_point_index_dict(discard_year=discard_year)
    
    @property
    @util.cache.file.decorator()
    def concentrations_same_points_except_year_dict(self):
        return super().concentrations_same_points_except_year_dict

    @property
    @util.cache.file.decorator()
    def sample_covariance_dict(self):
        return super().sample_covariance_dict
    
    @property
    @util.cache.file.decorator()
    def same_box_correlation_matrix_lower_triangle(self):
        return super().same_box_correlation_matrix_lower_triangle
    
    @property
    @util.cache.file.decorator()
    def different_boxes_quantity_lower_triangle_matrix(self):
        return super().different_boxes_quantity_lower_triangle_matrix
    
    @property
    @util.cache.file.decorator()
    def different_boxes_correlation_lower_triangle_matrix(self):
        return super().different_boxes_correlation_lower_triangle_matrix
    
    @property
    @util.cache.file.decorator()
    def correlation_matrix(self):
        return super().correlation_matrix
    
    
    ## cache files
    
    def map_indices_to_point_index_dict_cache_file(self, discard_year=False):
        return measurements.universal.constants.MAP_INDEX_TO_POINT_INDEX_DICT_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, discard_year=discard_year)
    
    def concentrations_same_points_except_year_dict_cache_file(self):
        return measurements.universal.constants.CONCENTRATIONS_SAME_POINTS_EXCEPT_YEAR_DICT_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements)
    
    def sample_covariance_dict_cache_file(self):
        return measurements.universal.constants.SAMPLE_COVARIANCE_DICT_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements, max_year_diff=self.max_year_diff)
    
    def same_box_correlation_matrix_lower_triangle_cache_file(self):
        return measurements.universal.constants.SAMPLE_CORRELATION_MATRIX_SAME_BOX_LOWER_TRIANGLE_MATRIX_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_abs_correlation=self.min_abs_correlation, standard_deviation_id=self.measurements.standard_deviation_id_without_sample_lsm, dtype=self.dtype, matrix_format=self.matrix_format)    
    
    def different_boxes_quantity_lower_triangle_matrix_cache_file(self):
        return measurements.universal.constants.SAMPLE_QUANTITY_MATRIX_DIFFERENT_BOXES_LOWER_TRIANGLE_MATRIX_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements, max_year_diff=self.max_year_diff, min_abs_correlation=self.min_abs_correlation, standard_deviation_id=self.measurements.standard_deviation_id_without_sample_lsm, dtype=self.dtype, matrix_format=self.matrix_format)
    
    def different_boxes_correlation_lower_triangle_matrix_cache_file(self):
        return measurements.universal.constants.SAMPLE_CORRELATION_MATRIX_DIFFERENT_BOXES_LOWER_TRIANGLE_MATRIX_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements, max_year_diff=self.max_year_diff, min_abs_correlation=self.min_abs_correlation, standard_deviation_id=self.measurements.standard_deviation_id_without_sample_lsm, dtype=self.dtype, matrix_format=self.matrix_format)
    
    def correlation_matrix_cache_file(self):
        return measurements.universal.constants.SAMPLE_CORRELATION_MATRIX_FILE.format(tracer=self.measurements.tracer, data_set=self.measurements.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements, max_year_diff=self.max_year_diff, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, standard_deviation_id=self.measurements.standard_deviation_id_without_sample_lsm, dtype=self.dtype, matrix_format=self.matrix_format)



import os.path
import datetime

from measurements.constants import BASE_DIR


## data information
DATA_DIR = os.path.join(BASE_DIR, 'dop', 'data')

LADOLFI_2002_DIR = os.path.join(DATA_DIR, 'Ladolfi2002')
LADOLFI_2002_MEASUREMENT_FILE = os.path.join(LADOLFI_2002_DIR, 'CD139_DOP_prepared.txt')
LADOLFI_2002_START_DATE = datetime.date(2002, 3, 1)
LADOLFI_2002_END_DATE = datetime.date(2002, 4, 15)
LADOLFI_2002_VALID_DATA_FLAG = 1

LADOLFI_2004_DIR = os.path.join(DATA_DIR, 'Ladolfi2004')
LADOLFI_2004_MEASUREMENT_FILE = os.path.join(LADOLFI_2004_DIR, 'D279_DOP_prepared.txt')
LADOLFI_2004_START_DATE = datetime.date(2004, 4, 4)
LADOLFI_2004_END_DATE = datetime.date(2004, 5, 10)
LADOLFI_2004_VALID_DATA_FLAG = 0

YOSHIMURA_2007_DIR = os.path.join(DATA_DIR, 'Yoshimura2007')
YOSHIMURA_2007_MEASUREMENT_FILE = os.path.join(YOSHIMURA_2007_DIR, 'Yoshimura2007.txt')

DATA_FILENAME = 'data.npy'

## sample lsm
from measurements.po4.wod.constants import SAMPLE_LSM

## deviation
DEVIATION_CONCENTRATION_NOISE_RATIO = 1.5







from util.logging import Logger

import measurements.po4.wod.data.values
import measurements.land_sea_mask.lsm
import measurements.util.plot

with Logger():
    lsm = measurements.land_sea_mask.lsm.LandSeaMaskWOA13R(t_dim=48)
    file = '/tmp/wod13_{}_value_histogram_{}_{}.png'.format('{}', lsm, '{}')
    min_measurements = 1000
    step_size = 0.05

    m = measurements.po4.wod.data.values.measurement_dict()
    measurements.util.plot.value_histograms(m, lsm=lsm, file=file.format('po4', '{}'), min_measurements=min_measurements, step_size=step_size)

    m = measurements.po4.wod.data.values.measurement_dict()
    m.set_min_result(0.01)
    m.log_results()
    measurements.util.plot.value_histograms(m, lsm=lsm, file=file.format('log_po4', '{}'), min_measurements=min_measurements, step_size=step_size)





import os.path
import numpy as np

from measurements.constants import BASE_DIR


## general
LSM_DIR = os.path.join(BASE_DIR, 'land_sea_masks')
LSM_NPY_FILENAME = 'land_sea_mask.npy'
DEPTH_NPY_FILENAME = 'depth_levels.npy'


## TMM (128x64x15)
TMM_DIR = os.path.join(LSM_DIR, 'TMM')
TMM_PETSC_FILE = os.path.join(TMM_DIR, 'land_sea_mask.petsc')
TMM_DEPTH_TXT_FILE = os.path.join(TMM_DIR, 'depth_levels.txt')
TMM_NORMALIZED_VOLUMES_PETSC_FILE = os.path.join(TMM_DIR, 'normalized_volumes.petsc')


## WOA13 (360x180x137)
WOA13_DIR = os.path.join(LSM_DIR, 'WOA13')
WOA13_LSM_TXT_FILE = os.path.join(WOA13_DIR, 'land_sea_mask.msk')
WOA13_DEPTH_TXT_FILE = os.path.join(WOA13_DIR, 'depth_levels.txt')




import numpy as np
import bisect
import logging

import measurements.land_sea_mask.lsm

import util.io.np
import util.io.netcdf
import util.io.fs
import util.math.interpolate


def load_from_netcdf(netcdf_file, netcdf_dataname):
    logging.debug('Loading data from {}.'.format(netcdf_file))

    data = util.io.netcdf.load(netcdf_file, netcdf_dataname)
    data = np.swapaxes(data, 1, 3)  # netcdf shape: (12, 15, 64, 128)

    return data


def save():
    from .constants import NOBS_NETCDF_ANNUAL_FILE, NOBS_NETCDF_MONTHLY_FILE, NOBS_NETCDF_DATANAME, NOBS_FILE
    from .constants import VARIS_NETCDF_ANNUAL_FILE, VARIS_NETCDF_MONTHLY_FILE, VARIS_NETCDF_DATANAME, VARIS_FILE
    from .constants import MEANS_NETCDF_ANNUAL_FILE, MEANS_NETCDF_MONTHLY_FILE, MEANS_NETCDF_DATANAME, MEANS_FILE
    from .constants import ANNUAL_THRESHOLD, VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR, VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR, VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND
    from measurements.po4.constants import DEVIATION_MIN_MEASUREMENTS, DEVIATION_MIN_VALUE


    ## concatenate annual and montly WOA data
    METOS_Z_LEFT = measurements.land_sea_mask.lsm.LandSeaMaskTMM().z_left
    
    z_index_annual_threshold = bisect.bisect_right(METOS_Z_LEFT, ANNUAL_THRESHOLD)
    logging.debug('Taking annual data from z index {}.'.format(z_index_annual_threshold))

    for (netcdf_annual_file, netcdf_monthly_file, netcdf_dataname, npy_file, divide) in ((NOBS_NETCDF_ANNUAL_FILE, NOBS_NETCDF_MONTHLY_FILE, NOBS_NETCDF_DATANAME, NOBS_FILE, True), (VARIS_NETCDF_ANNUAL_FILE, VARIS_NETCDF_MONTHLY_FILE, VARIS_NETCDF_DATANAME, VARIS_FILE, False), (MEANS_NETCDF_ANNUAL_FILE, MEANS_NETCDF_MONTHLY_FILE, MEANS_NETCDF_DATANAME, MEANS_FILE, False)):

        logging.debug('Preparing {}.'.format(npy_file))

        data_monthly = load_from_netcdf(netcdf_monthly_file, netcdf_dataname)
        data_annual = load_from_netcdf(netcdf_annual_file, netcdf_dataname)

        t_dim = data_monthly.shape[0]
        if divide:
            factor = 1 / t_dim
        else:
            factor = 1

        for t in range(t_dim):
            data_monthly[t,:,:,z_index_annual_threshold:] = data_annual[0,:,:,z_index_annual_threshold:] * factor

        util.io.np.save(npy_file, data_monthly, make_read_only=True, create_path_if_not_exists=True)


    ## revise variance
    nobs = np.load(NOBS_FILE)
    vari = np.load(VARIS_FILE)
    variance_min_value = DEVIATION_MIN_VALUE**2
    vari[vari < variance_min_value] = variance_min_value

    data_index_t, data_index_x, data_index_y, data_index_z = np.where(nobs >= DEVIATION_MIN_MEASUREMENTS)
    data_points_indices = (data_index_t, data_index_x, data_index_y, data_index_z)
    data_points = np.array(data_points_indices).swapaxes(0,1)
    data_values = vari[data_points_indices]

    assert data_points.ndim == 2
    assert data_values.ndim == 1
    assert len(data_points) == len(data_values)


    ## interpolate variance
    interpolation_index_t, interpolation_index_x, interpolation_index_y, interpolation_index_z = np.where(nobs < DEVIATION_MIN_MEASUREMENTS)
    interpolation_points_indices = (interpolation_index_t, interpolation_index_x, interpolation_index_y, interpolation_index_z)
    interpolation_points = np.array(interpolation_points_indices).swapaxes(0,1)

    logging.debug('Interpolating variance for {} points.'.format(interpolation_points.shape[0]))

    interpolator = util.math.interpolate.Periodic_Interpolator(data_points, data_values, point_range_size=vari.shape, wrap_around_amount=VARI_INTERPOLATION_AMOUNT_OF_WRAP_AROUND, number_of_linear_interpolators=VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR, single_overlapping_amount_linear_interpolators=VARI_INTERPOLATION_TOTAL_OVERLAPPING_OF_LINEAR_INTERPOLATOR/VARI_INTERPOLATION_NUMBER_OF_LINEAR_INTERPOLATOR)

    vari[interpolation_points_indices] = interpolator.interpolate(interpolation_points)

    assert np.all(np.isfinite(np.logical_not(np.isnan(vari))))


    ## saving interpolated variance
    logging.debug('Saving interpolated variance.')
    util.io.fs.make_writable(VARIS_FILE)
    util.io.np.save(VARIS_FILE, vari, make_read_only=True, create_path_if_not_exists=True)




import numpy as np

import logging
logger = logging.getLogger(__name__)

import measurements.dop.box.load
import measurements.po4.woa.data13.load
import util.io.np



def npy_or_save_dop_and_po4(npy_file, dop_function, po4_function):
    try:
        logger.debug('Loading data from {}.'.format(npy_file))

        data = np.load(npy_file)

    except (OSError, IOError):
        logger.debug('File {} does not exists. Calculating data.'.format(npy_file))

        dop = dop_function()
        po4 = po4_function()
        dop = dop.reshape((1,) + dop.shape)
        po4 = po4.reshape((1,) + po4.shape)

        data = np.append(dop, po4, axis=0)
        util.io.np.save(npy_file, data, make_read_only=True, create_path_if_not_exists=True)

    return data


def nobs():
    from .constants import NOBS_FILE
    data = npy_or_save_dop_and_po4(NOBS_FILE, measurements.dop.box.load.nobs, measurements.po4.woa.data13.load.nobs)

    assert np.all(data[np.logical_not(np.isnan(data))] >= 0)

    return data


def means():
    from .constants import MEANS_FILE
    data = npy_or_save_dop_and_po4(MEANS_FILE, measurements.dop.box.load.means, measurements.po4.woa.data13.load.means)

    assert np.all(data[np.logical_not(np.isnan(data))] >= 0)

    return data


def variances():
    from .constants import VARIS_FILE
    data = npy_or_save_dop_and_po4(VARIS_FILE, measurements.dop.box.load.variances, measurements.po4.woa.data13.load.variances)

    assert np.all(data[np.logical_not(np.isnan(data))] > 0)

    return data



#import overrides
import numpy as np

import measurements.universal.data
import measurements.universal.constants
import measurements.constants
import measurements.po4.constants
import measurements.po4.wod.cruise
import measurements.po4.wod.dict
import measurements.po4.wod.constants

import util.math.sort
import util.cache.file
import util.cache.memory
import util.logging

logger = util.logging.logger



## data load functions

@util.cache.file.decorator(cache_file_function=lambda :measurements.po4.wod.constants.CRUISES_LIST_FILE)
def cruises_list():
    cc = measurements.po4.wod.cruise.CruiseCollection()
    cc.load_cruises_from_netcdf_files(measurements.po4.wod.constants.CRUISES_DATA_DIR)
    return cc.cruises


@util.cache.file.decorator(cache_file_function=lambda :measurements.po4.wod.constants.MEASUREMENTS_DICT_UNSORTED_FILE)
def measurement_dict():
    m = measurements.po4.wod.dict.MeasurementsUnsortedDict()
    m.add_cruises(cruises_list())
    return m


@util.cache.file.decorator(cache_file_function=lambda :measurements.po4.wod.constants.POINTS_AND_RESULTS_FILE)
def points_and_results():
    logger.debug('Loading and calculating measurements.')

    ## load measurements
    m = measurement_dict()

    values = m.items()
    assert values.ndim == 2
    n = values.shape[1]
    assert n == 5

    ## sort measurements
    sorted_indices = util.math.sort.lex_sorted_indices(values)
    assert sorted_indices.ndim == 1
    values = values[sorted_indices]

    ## split measurements
    points = values[:, :-1]
    results = values[:, -1]

    return{'points': points, 'results': results}


def points():
    points_and_results()['points']

def results():
    points_and_results()['results']



## measurement classes

class Measurements(measurements.universal.data.MeasurementsAnnualPeriodicCache):

    def __init__(self, sample_t_dim=measurements.po4.wod.constants.SAMPLE_T_DIM, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):

        tracer = 'po4'
        data_set_name = 'wod_2013'

        sample_lsm = measurements.po4.wod.constants.SAMPLE_LSM
        sample_lsm.t_dim = sample_t_dim
        min_deviation = measurements.po4.constants.DEVIATION_MIN_VALUE

        super().__init__(sample_lsm, tracer=tracer, data_set_name=data_set_name, min_standard_deviation=min_deviation, min_measurements_correlation=min_measurements_correlation)

        self.fill_strategy = 'interpolate'

        try:
            INTERPOLATOR_OPTIONS = measurements.po4.wod.constants.INTERPOLATOR_OPTIONS
        except AttributeError:
            pass
        else:
            try:
                interpolator_option = INTERPOLATOR_OPTIONS['mean']['concentration'][measurements.constants.MEAN_MIN_MEASUREMENTS][str(sample_lsm)]
            except KeyError:
                pass
            else:
                self.set_interpolator_options('concentration_means', interpolator_option)
            try:
                interpolator_option = INTERPOLATOR_OPTIONS['deviation']['concentration'][measurements.constants.DEVIATION_MIN_MEASUREMENTS][str(sample_lsm)]
            except KeyError:
                pass
            else:
                self.set_interpolator_options('concentration_standard_deviations', interpolator_option)
            try:
                interpolator_option = INTERPOLATOR_OPTIONS['deviation']['average_noise'][measurements.constants.DEVIATION_MIN_MEASUREMENTS][str(sample_lsm)]
            except KeyError:
                pass
            else:
                self.set_interpolator_options('average_noise_standard_deviations', interpolator_option)


    @property
    @util.cache.memory.method_decorator()
    @util.cache.file.decorator()
    #@overrides.overrides
    def points(self):
        return points()

    @property
    @util.cache.memory.method_decorator()
    @util.cache.file.decorator()
    #@overrides.overrides
    def values(self):
        return results()



class MeasurementsNearWater(measurements.universal.data.MeasurementsAnnualPeriodicNearWaterCache):

    def __init__(self, water_lsm=None, max_box_distance_to_water=0, sample_t_dim=measurements.po4.wod.constants.SAMPLE_T_DIM, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):
        measurements = Measurements(sample_t_dim=sample_t_dim, min_measurements_correlation=min_measurements_correlation)
        super().__init__(measurements, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water)




import os.path

import numpy as np

from measurements.constants import BASE_DIR

## base dir

MEASUREMENT_DIR = os.path.join(BASE_DIR, '{tracer}', '{data_set}')


## data

DATA_DIR = os.path.join(MEASUREMENT_DIR, 'data')

POINTS_FILE = os.path.join(DATA_DIR, 'measurement_points.npy')
VALUES_FILE = os.path.join(DATA_DIR, 'measurement_values.npy')
MEASUREMENTS_DICT_FILE = os.path.join(DATA_DIR, 'measurements_dict.ppy')

SEPERATOR = '_-_'

NEAR_WATER_DATA_SET_NAME = SEPERATOR.join([
'{base_data_set_name}',
'{water_lsm}_water_{max_box_distance_to_water:d}'])

NEAR_WATER_PROJECTION_MASK_FILE = os.path.join(DATA_DIR, 'near_water_projection_matrix.{matrix_format}.npz')

INTERPOLATION_FILL_STRATEGY = 'interpolate_{scaling_values}_{interpolator_options}'


## mean

MEAN_MIN_MEASUREMENTS = 2

MEAN_DIR = os.path.join(MEASUREMENT_DIR, 'mean')

MEAN_FILE = os.path.join(MEAN_DIR, SEPERATOR.join([
'concentration_mean',
'for_{target}',
'sample_{sample_lsm}',
'min_values_{min_measurements:d}',
'fill_{fill_strategy}.npy']))

MEAN_ID = SEPERATOR.join([
'sample_{sample_lsm}',
'min_values_{min_measurements:d}'])


## deviation

DEVIATION_MIN_MEASUREMENTS = 3

DEVIATION_DIR = os.path.join(MEASUREMENT_DIR, 'deviation')

DEVIATION_FILE = os.path.join(DEVIATION_DIR, SEPERATOR.join([
'{deviation_type}',
'for_{target}',
'sample_{sample_lsm}',
'min_values_{min_measurements:d}',
'min_{min_standard_deviation:g}',
'fill_{fill_strategy}.npy']))

DEVIATION_ID = SEPERATOR.join([
'sample_{sample_lsm}',
'min_values_{min_measurements:d}',
'min_{min_standard_deviation:g}'])


## correlation

CORRELATION_MIN_MEASUREMENTS = 30
CORRELATION_MIN_ABS_VALUE = 0.01
CORRELATION_MAX_ABS_VALUE = 0.99
CORRELATION_CHOLESKY_MIN_DIAG_VALUE = 0.1
CORRELATION_CHOLESKY_ORDERING_METHOD = 'default'
CORRELATION_CHOLEKSY_REORDER_AFTER_EACH_STEP = True
CORRELATION_DTYPE = np.dtype(np.float32)
CORRELATION_FORMAT = 'csc'

# files

CORRELATION_ID = SEPERATOR.join([
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'min_abs_{min_abs_correlation}',
'max_abs_{max_abs_correlation}',
'ordering_{cholesky_ordering_method_correlation}_{cholesky_reordering_correlation:d}',
'min_diag_{cholesky_min_diag_value:.0e}',
'dev:_{standard_deviation_id}'])

CORRELATION_DIR = os.path.join(MEASUREMENT_DIR, 'correlation')

MAP_INDEX_TO_POINT_INDEX_DICT_FILE = os.path.join(CORRELATION_DIR,
'map_indices_to_point_index_dict', SEPERATOR.join([
'map_indices_to_point_index_dict',
'sample_{sample_lsm}',
'year_discarded_{discard_year}.ppy']))

CONCENTRATIONS_SAME_POINTS_EXCEPT_YEAR_DICT_FILE = os.path.join(CORRELATION_DIR,
'concentrations_same_points_except_year_dict', SEPERATOR.join([
'concentrations_same_points_except_year_dict', 
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}.ppy']))

SAMPLE_COVARIANCE_DICT_FILE = os.path.join(CORRELATION_DIR,
'sample_covariance', SEPERATOR.join([
'sample_covariance_dict.nonstationary',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'max_year_diff_{max_year_diff:0>2}.ppy']))

SAMPLE_CORRELATION_MATRIX_SAME_BOX_LOWER_TRIANGLE_MATRIX_FILE = os.path.join(CORRELATION_DIR,
'sample_correlation', SEPERATOR.join([
'sample_correlation.same_box.lower_triangle',
'sample_{sample_lsm}',
'min_abs_{min_abs_correlation}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))

SAMPLE_QUANTITY_MATRIX_DIFFERENT_BOXES_LOWER_TRIANGLE_MATRIX_FILE = os.path.join(CORRELATION_DIR,
'sample_quantity', SEPERATOR.join([
'sample_quantity.different_boxes.lower_triangle',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'max_year_diff_{max_year_diff:0>2}',
'min_abs_{min_abs_correlation}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))

SAMPLE_CORRELATION_MATRIX_DIFFERENT_BOXES_LOWER_TRIANGLE_MATRIX_FILE = os.path.join(CORRELATION_DIR,
'sample_correlation', SEPERATOR.join([
'sample_correlation.different_boxes.lower_triangle',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'max_year_diff_{max_year_diff:0>2}',
'min_abs_{min_abs_correlation}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))

SAMPLE_CORRELATION_MATRIX_FILE = os.path.join(CORRELATION_DIR,
'sample_correlation', SEPERATOR.join([
'sample_correlation',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'min_abs_{min_abs_correlation}',
'max_abs_{max_abs_correlation}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))

CORRELATION_MATRIX_POSITIVE_DEFINITE_FILE = os.path.join(CORRELATION_DIR,
'positive_definite', SEPERATOR.join([
'correlation',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'min_abs_{min_abs_correlation}',
'max_abs_{max_abs_correlation}',
'ordering_{cholesky_ordering_method_correlation}_{cholesky_reordering_correlation:d}',
'min_diag_{cholesky_min_diag_value:.0e}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))

CORRELATION_MATRIX_POSITIVE_DEFINITE_REDUCTION_FACTORS_FILE = os.path.join(CORRELATION_DIR,
'positive_definite', SEPERATOR.join([
'reduction_factors',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'min_abs_{min_abs_correlation}',
'max_abs_{max_abs_correlation}',
'ordering_{cholesky_ordering_method_correlation}_{cholesky_reordering_correlation:d}',
'min_diag_{cholesky_min_diag_value:.0e}',
'dev:_{standard_deviation_id}.npy']))

CORRELATION_MATRIX_CHOLESKY_FACTOR_FILE = os.path.join(CORRELATION_DIR,
'positive_definite', SEPERATOR.join([
'cholesky_{factor_type}',
'sample_{sample_lsm}',
'min_values_{min_measurements_correlation:0>2d}',
'min_abs_{min_abs_correlation}',
'max_abs_{max_abs_correlation}',
'ordering_{cholesky_ordering_method_correlation}_{cholesky_reordering_correlation:d}',
'min_diag_{cholesky_min_diag_value:.0e}',
'dev:_{standard_deviation_id}',
'{dtype}.{matrix_format}.npz']))



import abc
import os.path

import numpy as np
import scipy.sparse
#import overrides

import measurements.universal.dict
import measurements.universal.interpolate
import measurements.universal.sample_data
import measurements.universal.constants

import util.cache.file
import util.cache.memory
import util.options
import util.str
import util.logging

logger = util.logging.logger



class Measurements():

    def __init__(self, tracer=None, data_set_name=None):
        self._tracer = tracer
        self._data_set_name = data_set_name
        self.cholesky_ordering_method_correlation = 'best'
        logger.debug('{}: initialized with tracer {} and data set {}.'.format(self.__class__.__name__, tracer, data_set_name))


    def __repr__(self):
        return '<measurements for tracer "{tracer}" with data set "{data_set_name}" and {number_of_measurements} measurements>'.format(tracer=self.tracer, data_set_name=self.data_set_name, number_of_measurements=self.number_of_measurements)

    def __str__(self):
        return '{tracer}:{data_set_name}'.format(tracer=self.tracer, data_set_name=self.data_set_name)


    @property
    def tracer(self):
        if self._tracer is not None:
            return self._tracer
        else:
            return ValueError('Tracer is not set.')

    @property
    def data_set_name(self):
        if self._data_set_name is not None:
            return self._data_set_name
        else:
            return ValueError('Data set name is not set.')


    @property
    @abc.abstractmethod
    def points(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def values(self):
        raise NotImplementedError()


    @property
    def number_of_measurements(self):
        return len(self.points)

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    def measurements_dict(self):
        m = measurements.universal.dict.MeasurementsDict()
        m.append_values(self.points, self.values)
        return m


    @property
    @abc.abstractmethod
    def means(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def standard_deviations(self):
        raise NotImplementedError()

    @property
    def variances(self):
        return self.standard_deviations ** 2


    @property
    def correlations_own(self):
        return scipy.sparse.eye(self.number_of_measurements)

    @property
    def correlations_own_cholesky_decomposition(self):
        import util.math.sparse.decompose.with_cholmod
        P, L = util.math.sparse.decompose.with_cholmod.cholesky(self.correlations_own, ordering_method=self.cholesky_ordering_method_correlation, return_type=util.math.sparse.decompose.with_cholmod.RETURN_P_L)
        return {'P': P, 'L': L}

    def correlations_other(self, measurements=None):
        return scipy.sparse.dia_matrix((self.number_of_measurements, measurements.number_of_measurements))


    def correlations(self, measurements=None):
        if measurements is None or measurements == self:
            return self.correlations_own
        else:
            return self.correlations_other(measurements=measurements)




class MeasurementsAnnualPeriodicBase(Measurements):

    def __init__(self, sample_lsm, tracer=None, data_set_name=None, min_standard_deviation=np.finfo(np.float).resolution, min_abs_correlation=measurements.universal.constants.CORRELATION_MIN_ABS_VALUE, max_abs_correlation=measurements.universal.constants.CORRELATION_MAX_ABS_VALUE, min_measurements_mean=measurements.universal.constants.MEAN_MIN_MEASUREMENTS, min_measurements_standard_deviation=measurements.universal.constants.DEVIATION_MIN_MEASUREMENTS, min_measurements_correlation=measurements.universal.constants.CORRELATION_MIN_MEASUREMENTS):

        super().__init__(tracer=tracer, data_set_name=data_set_name)

        self._sample_lsm = sample_lsm

        self.min_measurements_mean = min_measurements_mean

        self.min_measurements_standard_deviation = min_measurements_standard_deviation
        self.min_standard_deviation = min_standard_deviation

        self.min_measurements_correlation = min_measurements_correlation
        self.min_abs_correlation = min_abs_correlation
        self.max_abs_correlation = max_abs_correlation

        self.cholesky_min_diag_value_correlation = measurements.universal.constants.CORRELATION_CHOLESKY_MIN_DIAG_VALUE
        self.cholesky_ordering_method_correlation = measurements.universal.constants.CORRELATION_CHOLESKY_ORDERING_METHOD
        self.cholesky_reordering_correlation = measurements.universal.constants.CORRELATION_CHOLEKSY_REORDER_AFTER_EACH_STEP
        self.matrix_format_correlation = measurements.universal.constants.CORRELATION_FORMAT


    @property
    def dtype_correlation(self):
        return measurements.universal.constants.CORRELATION_DTYPE


    ## general sample data

    @property
    def sample_lsm(self):
        return self._sample_lsm

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    def _sample_mean_and_deviation(self):
        return measurements.universal.sample_data.SampleMeanAndDeviation(self.points, self.values, self.sample_lsm)


    ## mean

    @property
    def sample_means(self):
        return self._sample_mean_and_deviation.sample_concentration_means(min_measurements=self.min_measurements_mean)

    @property
    #@overrides.overrides
    def means(self):
        data = self.sample_means
        if data.count() == len(data):
            return data.data
        else:
            raise TooFewValuesError('It was not possible to calculate all values from the sample values, because to few sample values are available.')


    ## deviation

    @property
    def sample_concentration_standard_deviations(self):
        return self._sample_mean_and_deviation.sample_concentration_standard_deviations(min_measurements=self.min_measurements_standard_deviation, min_value=0)

    @property
    def concentration_standard_deviations(self):
        data = self.sample_concentration_standard_deviations
        if data.count() == len(data):
            return data.data
        else:
            raise TooFewValuesError('It was not possible to calculate all values from the sample values, because to few sample values are available.')


    @property
    def sample_average_noise_standard_deviations(self):
        return self._sample_mean_and_deviation.sample_average_noise_standard_deviations(min_measurements=self.min_measurements_standard_deviation, min_value=self.min_standard_deviation)

    @property
    def average_noise_standard_deviations(self):
        data = self.sample_average_noise_standard_deviations
        if data.count() == len(data):
            return data.data
        else:
            raise TooFewValuesError('It was not possible to calculate all values from the sample values, because to few sample values are available.')


    @property
    def sample_noise_standard_deviations(self):
        return self._sample_mean_and_deviation.sample_noise_standard_deviations(min_measurements=self.min_measurements_standard_deviation, min_value=self.min_standard_deviation)

    @property
    def noise_standard_deviations(self):
        data = self.sample_noise_standard_deviations
        if data.count() < len(data):
            data[data.mask] = self.average_noise_standard_deviations[data.mask]
        assert data.count() == len(data)
        return data.data


    @property
    #@overrides.overrides
    def standard_deviations(self):
        standard_deviations = (self.concentration_standard_deviations**2 + self.noise_standard_deviations**2)**(1/2)
        return standard_deviations


    ## correlation

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.min_measurements_correlation', 'self.min_abs_correlation', 'self.max_abs_correlation', 'self.matrix_format_correlation', 'self.dtype_correlation'))
    def _sample_correlation(self):
        return measurements.universal.sample_data.SampleCorrelationMatrix(self, self.sample_lsm, self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, matrix_format=self.matrix_format_correlation, dtype=self.dtype_correlation)


    @property
    def correlations_own_sample_matrix(self):
        return self._sample_correlation.correlation_matrix


    @property
    #@overrides.overrides
    def correlations_own(self):
        import util.math.sparse.decompose.with_cholmod
        correlation_matrix, reduction_factors = util.math.sparse.decompose.with_cholmod.approximate_positive_definite(self.correlations_own_sample_matrix, min_abs_value=self.min_abs_correlation, min_diag_value=self.cholesky_min_diag_value_correlation, ordering_method=self.cholesky_ordering_method_correlation, reorder_after_each_step=self.cholesky_reordering_correlation)
        return correlation_matrix.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)


    @property
    #@overrides.overrides
    def correlations_own_cholesky_decomposition(self):
        import util.math.sparse.decompose.with_cholmod
        P, L = util.math.sparse.decompose.with_cholmod.cholesky(self.correlations_own, ordering_method=self.cholesky_ordering_method_correlation, return_type=util.math.sparse.decompose.with_cholmod.RETURN_P_L)
        return {'P': P.asformat(self.matrix_format_correlation).astype(np.int8), 'L': L.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)}




class MeasurementsAnnualPeriodic(MeasurementsAnnualPeriodicBase):

    def __init__(self, *args, **kargs):

        super().__init__(*args, **kargs)

        self.POSSIBLE_FILL_STRATEGIES = ('auto', 'point_average', 'lsm_average', 'constant', 'interpolate')
        self.POSSIBLE_KINDS = ('concentration_means', 'concentration_standard_deviations', 'average_noise_standard_deviations')
        self._interpolator_options = {}
        self._constant_fill_values = {}


    ## interpolater

    def _check_kind(self, kind):
        if kind not in self.POSSIBLE_KINDS:
            raise ValueError('Kind must be in {}, but it is {}.'.format(self.POSSIBLE_KINDS, kind))
        return kind


    def get_interpolator_options(self, kind):
        try:
            return self._interpolator_options[kind]
        except KeyError:
            return (1,1,0,0)

    def set_interpolator_options(self, kind, value):
        self._interpolator_options[self._check_kind(kind)] = value


    @property
    def interpolator_scaling_values(self):
        try:
            return self._interpolator_scaling_values
        except AttributeError:
            return measurements.universal.interpolate.default_scaling_values(self.sample_lsm)

    @interpolator_scaling_values.setter
    def interpolator_scaling_values(self, value):
        self._interpolator_scaling_values = value


    @property
    def _interpolator(self):
        return measurements.universal.interpolate.Interpolator_Annual_Periodic(self.sample_lsm, scaling_values=self.interpolator_scaling_values)


    ## fill strategy

    @property
    def fill_strategy(self):
        try:
            return self._fill_strategy
        except AttributeError:
            return 'auto'

    @fill_strategy.setter
    def fill_strategy(self, value):
        if value in self.POSSIBLE_FILL_STRATEGIES:
            self._fill_strategy = value
        else:
            raise ValueError('Fill strategy {} is unknown. Only fill strategies {} are supported.'.format(value, self.POSSIBLE_FILL_STRATEGIES))


    def get_constant_fill_value(self, kind):
        try:
            return self._constant_fill_values[kind]
        except KeyError:
            raise ValueError('Constant fill value is not set for kind {}.'.format(kind))

    def set_constant_fill_value(self, kind, value):
        self._constant_fill_values[self._check_kind(kind)] = value


    def _choose_fill_strategy(self, number_of_sample_values):
        number_of_lsm_values = self.sample_lsm.number_of_map_indices
        sample_data_fill_amount = number_of_sample_values / number_of_lsm_values
        if sample_data_fill_amount >= 0.001:
            fill_strategy = 'interpolate'
        elif number_of_sample_values >= 1:
            fill_strategy = 'point_average'
        else:
            fill_strategy = 'constant'

        logger.debug('{}: Chosen {} as fill strategy, since {:d} sample data are for {:%} of the sampe lsm available.'.format(self.__class__.__name__, fill_strategy, number_of_sample_values, sample_data_fill_amount))
        return fill_strategy


    def _fill_strategy_for_kind(self, kind):
        ## choose fill method
        fill_strategy = self.fill_strategy
        if fill_strategy == 'auto':
            number_of_sample_values = len(self._data_map_indices_dict(kind))
            fill_strategy = self._choose_fill_strategy(number_of_sample_values)

        ## return
        logger.debug('{}: Fill startegy to use is {}.'.format(self.__class__.__name__, fill_strategy))
        return fill_strategy


    def _fill_strategy_with_number_of_sample_values(self, number_of_sample_values):
        ## choose fill method
        fill_strategy = self.fill_strategy
        if fill_strategy == 'auto':
            fill_strategy = self._choose_fill_strategy(number_of_sample_values)

        ## check number of available sample values
        if number_of_sample_values == 0 and fill_strategy != 'constant':
            raise TooFewValuesError('No sample values are available. Fill method {} is not applicable.'.format(fill_strategy))

        ## return
        logger.debug('{}: Fill startegy to use is {}.'.format(self.__class__.__name__, fill_strategy))
        return fill_strategy


    ## data general

    def _data_map_indices_dict(self, kind):
        if kind == 'concentration_means':
            data_map_indices_dict = self._sample_mean_and_deviation.sample_concentration_means_map_indices_dict(min_measurements=self.min_measurements_mean)
        elif kind == 'concentration_standard_deviations':
            data_map_indices_dict = self._sample_mean_and_deviation.sample_concentration_standard_deviations_map_indices_dict(min_measurements=self.min_measurements_standard_deviation, min_value=0)
        elif kind == 'average_noise_standard_deviations':
            data_map_indices_dict = self._sample_mean_and_deviation.sample_average_noise_standard_deviations_map_indices_dict(min_measurements=self.min_measurements_standard_deviation, min_value=self.min_standard_deviation)
        else:
            raise ValueError('Unknown kind {}.'.format(kind))
        return data_map_indices_dict


    ## data for sample lsm

    def _data_for_sample_lsm(self, kind):
        logger.debug('{}: Calculating {} data for sample lsm.'.format(self.__class__.__name__, kind))

        ## get data
        data_map_indices_dict = self._data_map_indices_dict(kind)
        map_indices_and_values = data_map_indices_dict.toarray()

        ## choose fill strategy
        fill_strategy = self._fill_strategy_with_number_of_sample_values(len(map_indices_and_values))
        logger.debug('{}: Filling sample lsm values with fill strategy {}.'.format(self.__class__.__name__, fill_strategy))

        ## apply fill_strategy
        if fill_strategy in ('point_average', 'lsm_average', 'constant'):

            if fill_strategy == 'point_average':
                sample_values = self._sample_mean_and_deviation._convert_map_indices_dict_to_array_for_points(data_map_indices_dict, is_discard_year=True)
                fill_value = sample_values.mean()
            elif fill_strategy == 'lsm_average':
                fill_value = map_indices_and_values[:,-1].mean()
            elif fill_strategy == 'constant':
                fill_value = self.get_constant_fill_value(kind)

            lsm_values = self.sample_lsm.insert_index_values_in_map(map_indices_and_values, no_data_value=fill_value, skip_values_on_land=False)

        elif fill_strategy == 'interpolate':
            interpolator_options = self.get_interpolator_options(kind)
            lsm_values = self._interpolator.interpolate_data_for_sample_lsm_with_map_indices(map_indices_and_values, interpolator_options)

        else:
            raise ValueError('Unknown fill method {}.'.format(fill_strategy))

        return lsm_values

    @property
    def means_for_sample_lsm(self):
        return self._data_for_sample_lsm('means')

    @property
    def concentration_standard_deviations_for_sample_lsm(self):
        return self._data_for_sample_lsm('concentration_standard_deviations')

    @property
    def average_noise_standard_deviations_for_sample_lsm(self):
        return self._data_for_sample_lsm('average_noise_standard_deviations')

    @property
    def standard_deviations_for_sample_lsm(self):
        return (self.concentration_standard_deviations_for_sample_lsm**2 + self.average_noise_standard_deviations_for_sample_lsm**2)**(1/2)


    ## data for sample points

    def _data_for_sample_points(self, kind):
        logger.debug('{}: Calculating {} data for sample points.'.format(self.__class__.__name__, kind))

        ## get data
        data_map_indices_dict = self._data_map_indices_dict(kind)
        data = self._sample_mean_and_deviation._convert_map_indices_dict_to_array_for_points(data_map_indices_dict, is_discard_year=True)
        number_of_values = data.count()
        number_of_points = len(data)

        logger.debug('{}: Got values for {:d} of {:d} points with sample data.'.format(self.__class__.__name__, number_of_values, number_of_points))

        ## fill if empty values
        if number_of_values < number_of_points:
            ## choose fill strategy
            fill_strategy = self._fill_strategy_with_number_of_sample_values(len(data_map_indices_dict))
            logger.debug('{}: Filling remaining {:%} sample points values with fill strategy {}.'.format(self.__class__.__name__, 1-number_of_values/number_of_points, fill_strategy))

            ## fill
            if fill_strategy == 'point_average':
                data[data.mask] = data.mean()
            elif fill_strategy == 'lsm_average':
                data[data.mask] = data_map_indices_dict.values().mean()
            elif fill_strategy == 'interpolate':
                data[data.mask] = self._interpolator.interpolate_data_for_points_from_interpolated_lsm_data(self._data_for_sample_lsm(kind), self.points[data.mask])
            elif fill_strategy == 'constant':
                data[data.mask] = self.get_constant_fill_value(kind)
            else:
                raise ValueError('Unknown fill method {}.'.format(fill_strategy))

        assert data.count() == len(data)
        return data.data


    @property
    #@overrides.overrides
    def means(self):
        return self._data_for_sample_points('concentration_means')

    @property
    #@overrides.overrides
    def concentration_standard_deviations(self):
        return self._data_for_sample_points('concentration_standard_deviations')

    @property
    #@overrides.overrides
    def average_noise_standard_deviations(self):
        return self._data_for_sample_points('average_noise_standard_deviations')




class MeasurementsNearWater(Measurements):

    def __init__(self, base_measurements, water_lsm=None, max_box_distance_to_water=0):
        self.base_measurements = base_measurements
        self.max_box_distance_to_water = max_box_distance_to_water

        if water_lsm is None:
            water_lsm = self.base_measurements.sample_lsm
        water_lsm = water_lsm.copy()
        water_lsm.t_dim = 0
        self.water_lsm = water_lsm

        Measurements.__init__(self)


    ## properties
    @property
    def tracer(self):
        return self.base_measurements.tracer

    @property
    def data_set_name(self):
        return measurements.universal.constants.NEAR_WATER_DATA_SET_NAME.format(base_data_set_name=self.base_measurements.data_set_name, water_lsm=self.water_lsm, max_box_distance_to_water=self.max_box_distance_to_water)


    ## projection methods

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    def near_water_projection_matrix(self):
        mask = self.water_lsm.coordinates_near_water_mask(self.base_measurements.points, max_box_distance_to_water=self.max_box_distance_to_water)

        n = mask.sum()
        m = len(mask)
        assert n <= m
        near_water_matrix = scipy.sparse.dok_matrix((n, m), dtype=np.int16)

        i = 0
        for j in range(m):
            if mask[j]:
                near_water_matrix[i, j] = 1
                i = i + 1
        assert i == n

        return near_water_matrix.tocsc()


    ## other methods

    @property
    #@overrides.overrides
    def points(self):
        return self.near_water_projection_matrix * self.base_measurements.points

    @property
    #@overrides.overrides
    def values(self):
        return self.near_water_projection_matrix * self.base_measurements.values

    @property
    #@overrides.overrides
    def means(self):
        return self.near_water_projection_matrix * self.base_measurements.means

    @property
    #@overrides.overrides
    def standard_deviations(self):
        return self.near_water_projection_matrix * self.base_measurements.standard_deviations

    @property
    #@overrides.overrides
    def correlations_own(self):
        return self.near_water_projection_matrix * self.base_measurements.correlations_own * self.near_water_projection_matrix.T

    #@overrides.overrides
    def correlations_other(self, measurements=None):
        return self.near_water_projection_matrix * self.base_measurements.correlations_other(measurements=measurements)




class MeasurementsAnnualPeriodicNearWater(MeasurementsNearWater, MeasurementsAnnualPeriodic):

    def __init__(self, base_measurements, water_lsm=None, max_box_distance_to_water=0):
        MeasurementsAnnualPeriodicBase.__init__(self, base_measurements.sample_lsm, min_standard_deviation=base_measurements.min_standard_deviation, min_abs_correlation=base_measurements.min_abs_correlation, max_abs_correlation=base_measurements.max_abs_correlation, min_measurements_mean=base_measurements.min_measurements_mean, min_measurements_standard_deviation=base_measurements.min_measurements_standard_deviation, min_measurements_correlation=base_measurements.min_measurements_correlation)
        super().__init__(base_measurements, water_lsm=water_lsm, max_box_distance_to_water=max_box_distance_to_water)


    @property
    #@overrides.overrides
    def concentration_standard_deviations(self):
        return self.near_water_projection_matrix * self.base_measurements.concentration_standard_deviations

    @property
    #@overrides.overrides
    def noise_standard_deviations(self):
        return self.near_water_projection_matrix * self.base_measurements.noise_standard_deviations

    @property
    #@overrides.overrides
    def average_noise_standard_deviations(self):
        return self.near_water_projection_matrix * self.base_measurements.average_noise_standard_deviations


    @property
    #@overrides.overrides
    def correlations_own_sample_matrix(self):
        return self.near_water_projection_matrix * self.base_measurements.correlations_own_sample_matrix * self.near_water_projection_matrix.T

    @property
    #@overrides.overrides
    def correlations_own(self):
        return MeasurementsAnnualPeriodicBase.correlations_own




class MeasurementsAnnualPeriodicUnion(MeasurementsAnnualPeriodic):

    def __init__(self, *measurements_list):
        logger.debug('Initiating {} with measurements {}.'.format(self.__class__.__name__, measurements_list))

        if len(measurements_list) == 0:
            raise ValueError('{} must be initiated with at least one measurement object.'.format(self.__class__.__name__))

        ## sort
        measurements_list = sorted(measurements_list, key=lambda measurement: measurement.data_set_name)

        ## check same tracer and data set name
        n = len(measurements_list)
        for i in range(n-1):
            if measurements_list[i].tracer != measurements_list[i+1].tracer:
                raise ValueError('Measurements objects with different tracers ({} and {}) are not allowed!'.format(measurements_list[i].tracer, measurements_list[i+1].tracer))
            if measurements_list[i].data_set_name == measurements_list[i+1].data_set_name:
                raise ValueError('Measurements objects with same tracer ({}) and same data set name ({}) are not allowed!'.format(measurements_list[i].tracer, measurements_list[i].data_set_name))
            if measurements_list[i].sample_lsm != measurements_list[i+1].sample_lsm:
                raise ValueError('Measurements objects with different sample lsm ({} and {}) are not allowed!'.format(measurements_list[i].sample_lsm, measurements_list[i+1].sample_lsm))

        ## store
        self.measurements_list = measurements_list

        ## chose values for union
        sample_lsm = measurements_list[0].sample_lsm
        tracer = measurements_list[0].tracer
        data_set_name = ','.join(map(lambda measurement: measurement.data_set_name, measurements_list))

        def get_and_set_default_value(value_name, reduce_function):
            default_value = reduce_function(map(lambda measurement: getattr(measurement, value_name), measurements_list))
            for measurement in measurements_list:
                setattr(measurement, value_name, default_value)
            return default_value

        min_measurements_mean = get_and_set_default_value('min_measurements_mean', min)
        min_standard_deviation = get_and_set_default_value('min_standard_deviation', min)
        min_measurements_standard_deviation = get_and_set_default_value('min_measurements_standard_deviation', min)
        min_abs_correlation = get_and_set_default_value('min_abs_correlation', min)
        max_abs_correlation = get_and_set_default_value('max_abs_correlation', max)
        min_measurements_correlation = get_and_set_default_value('min_measurements_correlation', min)

        ## call super init
        super().__init__(sample_lsm, tracer=tracer, data_set_name=data_set_name, min_measurements_mean=min_measurements_mean, min_standard_deviation=min_standard_deviation, min_measurements_standard_deviation=min_measurements_standard_deviation, min_abs_correlation=min_abs_correlation, max_abs_correlation=max_abs_correlation, min_measurements_correlation=min_measurements_correlation)


    @property
    #@overrides.overrides
    def points(self):
        return np.concatenate(tuple(map(lambda measurement: measurement.points, self.measurements_list)), axis=0)

    @property
    #@overrides.overrides
    def values(self):
        return np.concatenate(tuple(map(lambda measurement: measurement.values, self.measurements_list)))

    @property
    #@overrides.overrides
    def number_of_measurements(self):
        return sum(map(lambda measurement: measurement.number_of_measurements, self.measurements_list))




class MeasurementsCollection(Measurements):

    def __init__(self, *measurements_list):
        logger.debug('Initiating {} with measurements {}.'.format(self.__class__.__name__, measurements_list))

        if len(measurements_list) == 0:
            raise ValueError('There are no measurements in the measurements list!')

        ## sort
        measurements_list = sorted(measurements_list, key=lambda measurement: measurement.data_set_name)
        measurements_list = sorted(measurements_list, key=lambda measurement: measurement.tracer)

        ## check same tracer and data set name
        n = len(measurements_list)
        for i in range(n-1):
            if measurements_list[i].tracer == measurements_list[i+1].tracer and measurements_list[i].data_set_name == measurements_list[i+1].data_set_name:
                raise ValueError('There is more then one measurements object with tracer {} and data set name {}!'.format(measurements_list[i].tracer, measurements_list[i].data_set_name))

        ## store
        self._measurements_list = measurements_list

        ## make tracer and data set name for collection
        tracer = ','.join(map(lambda measurement: measurement.tracer, measurements_list))
        data_set_name = ','.join(map(lambda measurement: measurement.data_set_name, measurements_list))
        if len(measurements_list) > 1:
            data_set_name = ','.join(map(lambda measurement: '{tracer}:{data_set_name}'.format(tracer=measurement.tracer, data_set_name=measurement.data_set_name), measurements_list))
        else:
            data_set_name = measurements_list[0].data_set_name
        super().__init__(tracer=tracer, data_set_name=data_set_name)

        ## correlation constants
        self.min_abs_correlation = min([measurement.min_abs_correlation for measurement in measurements_list])
        self.cholesky_min_diag_value_correlation = min([measurement.cholesky_min_diag_value_correlation for measurement in measurements_list])
        self.cholesky_ordering_method_correlation = measurements.universal.constants.CORRELATION_CHOLESKY_ORDERING_METHOD
        self.cholesky_reordering_correlation = measurements.universal.constants.CORRELATION_CHOLEKSY_REORDER_AFTER_EACH_STEP
        self.matrix_format_correlation = measurements.universal.constants.CORRELATION_FORMAT
        self.dtype_correlation = measurements.universal.constants.CORRELATION_DTYPE


    @property
    def measurements_list(self):
        return self._measurements_list


    def __str__(self):
        return ','.join(map(str, self.measurements_list))


    def __iter__(self):
        return self.measurements_list.__iter__()


    @property
    #@overrides.overrides
    def points(self):
        return np.concatenate(tuple(map(lambda measurement: measurement.points, self.measurements_list)), axis=0)

    @property
    #@overrides.overrides
    def values(self):
        return np.concatenate(tuple(map(lambda measurement: measurement.values, self.measurements_list)))


    @property
    #@overrides.overrides
    def number_of_measurements(self):
        return sum(map(lambda measurement: measurement.number_of_measurements, self.measurements_list))

    @property
    #@overrides.overrides
    def means(self):
        values = np.concatenate(tuple(map(lambda measurement: measurement.means, self.measurements_list)))
        assert len(values) == self.number_of_measurements
        return values

    @property
    #@overrides.overrides
    def standard_deviations(self):
        values = np.concatenate(tuple(map(lambda measurement: measurement.standard_deviations, self.measurements_list)))
        assert len(values) == self.number_of_measurements
        return values


    @property
    def correlations_own_sample_matrix(self):
        n = len(self.measurements_list)
        correlations = np.empty([n,n], dtype=object)

        for i in range(n):
            measurements_i = self.measurements_list[i]
            correlations[i, i] = measurements_i.correlations()
            for j in range(i+1, n):
                measurements_j = self.measurements_list[j]
                correlations[i, j] = measurements_i.correlations(measurements_j)
                correlations[j, i] = correlations[i, j].T

        correlations = scipy.sparse.bmat(correlations, format=self.matrix_format_correlation, dtype=self.dtype_correlation)
        assert correlations.shape == (self.number_of_measurements, self.number_of_measurements)
        return correlations


    @property
    #@overrides.overrides
    def correlations_own(self):
        import util.math.sparse.decompose.with_cholmod
        correlation_matrix, reduction_factors = util.math.sparse.decompose.with_cholmod.approximate_positive_definite(self.correlations_own_sample_matrix, min_abs_value=self.min_abs_correlation, min_diag_value=self.cholesky_min_diag_value_correlation, ordering_method=self.cholesky_ordering_method_correlation, reorder_after_each_step=self.cholesky_reordering_correlation)
        correlation_matrix = correlation_matrix.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)
        assert correlation_matrix.shape == (self.number_of_measurements, self.number_of_measurements)
        return correlation_matrix


    @property
    #@overrides.overrides
    def correlations_own_cholesky_decomposition(self):
        import util.math.sparse.decompose.with_cholmod
        P, L = util.math.sparse.decompose.with_cholmod.cholesky(self.correlations_own, ordering_method=self.cholesky_ordering_method_correlation, return_type=util.math.sparse.decompose.with_cholmod.RETURN_P_L)
        return {'P': P.asformat(self.matrix_format_correlation).astype(np.int8), 'L': L.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)}


    def _measurements_dict(self, convert_function=None):
        if convert_function is None:
            convert_function = lambda x: x

        results = {}

        for measurement in self.measurements_list:
            tracer = measurement.tracer
            data_set_name = measurement.data_set_name

            try:
                results[tracer]
            except KeyError:
                results[tracer] = {}

            results[tracer][data_set_name] = convert_function(measurement)

        return results


    @property
    def points_dict(self):
        return self._measurements_dict(convert_function=lambda m: m.points)


    def convert_measurements_dict_to_array(self, measurements_dict):
        value_list = [measurements_dict[measurement.tracer][measurement.data_set_name] for measurement in self.measurements_list]
        return np.concatenate(value_list)


    def subset(self, tracers):
        measurements_list = [measurement for measurement in self.measurements_list if measurement.tracer in tracers]
        subset = type(self)(*measurements_list)
        return subset




## caches

class MeasurementsCache():

    @property
    def mean_id(self):
        return ''

    @property
    def standard_deviation_id(self):
        return ''

    @property
    def correlation_id(self):
        return ''




class MeasurementsAnnualPeriodicBaseCache(MeasurementsCache, MeasurementsAnnualPeriodicBase):

    ## ids

    @property
    #@overrides.overrides
    def mean_id(self):
        return measurements.universal.constants.MEAN_ID.format(sample_lsm=self.sample_lsm, min_measurements=self.min_measurements_mean)

    @property
    #@overrides.overrides
    def standard_deviation_id(self):
        return measurements.universal.constants.DEVIATION_ID.format(sample_lsm=self.sample_lsm, min_measurements=self.min_measurements_standard_deviation, min_standard_deviation=self.min_standard_deviation)

    @property
    def standard_deviation_id_without_sample_lsm(self):
        seperator = measurements.universal.constants.SEPERATOR
        standard_deviation_id = seperator.join(self.standard_deviation_id.split(seperator)[1:])
        return standard_deviation_id

    @property
    #@overrides.overrides
    def correlation_id(self):
        return measurements.universal.constants.CORRELATION_ID.format(sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, cholesky_ordering_method_correlation=self.cholesky_ordering_method_correlation, cholesky_reordering_correlation=self.cholesky_reordering_correlation, cholesky_min_diag_value=self.cholesky_min_diag_value_correlation, standard_deviation_id=self.standard_deviation_id_without_sample_lsm)




class MeasurementsAnnualPeriodicCache(MeasurementsAnnualPeriodicBaseCache, MeasurementsAnnualPeriodic):

    ## ids

    def _fill_strategy_id(self, kind):
        ## if standrad deviation, use str for concentration and average_noise
        if kind == 'standard_deviations':
            concentration_fill_strategy = self._fill_strategy_id('concentration_standard_deviations')
            average_noise_fill_strategy = self._fill_strategy_id('average_noise_standard_deviations')
            ## if same for both, use only once
            if concentration_fill_strategy == average_noise_fill_strategy:
                fill_strategy = concentration_fill_strategy
            ## if both use interpolation, use only once but two interpolator_options
            elif concentration_fill_strategy.startswith('interpolate') and average_noise_fill_strategy.startswith('interpolate'):
                interpolator_options = '+'.join([','.join(map(str, self.get_interpolator_options(kind))) for kind in ('concentration_standard_deviations', 'average_noise_standard_deviations')])
                fill_strategy = measurements.universal.constants.INTERPOLATION_FILL_STRATEGY.format(scaling_values=','.join(map(str, self.interpolator_scaling_values)), interpolator_options=interpolator_options)
            ## if different strategies, append both strategies
            else:
                fill_strategy = util.str.merge([concentration_fill_strategy, average_noise_fill_strategy])
        ## else, get used fill strategy
        else:
            if kind == 'noise_standard_deviations':
                fill_strategy = self._fill_strategy_for_kind('average_' + kind)
            else:
                fill_strategy = self._fill_strategy_for_kind(kind)

            ## if interpolation, append options for interpolations
            if fill_strategy == 'interpolate':
                fill_strategy = measurements.universal.constants.INTERPOLATION_FILL_STRATEGY.format(scaling_values=','.join(map(str, self.interpolator_scaling_values)), interpolator_options=','.join(map(str, self.get_interpolator_options(kind))))

        return fill_strategy


    @property
    #@overrides.overrides
    def mean_id(self):
        return super().mean_id + '_-_fill_' + self._fill_strategy_id('concentration_means')

    @property
    #@overrides.overrides
    def standard_deviation_id(self):
        return super().standard_deviation_id + '_-_fill_' + self._fill_strategy_id('standard_deviations')



    ## points and values cache files

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def points(self):
        return super().points

    def points_cache_file(self):
        return measurements.universal.constants.POINTS_FILE.format(tracer=self.tracer, data_set=self.data_set_name)


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def values(self):
        return super().values

    def values_cache_file(self):
        return measurements.universal.constants.VALUES_FILE.format(tracer=self.tracer, data_set=self.data_set_name)


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def measurements_dict(self):
        return super().measurements_dict

    def measurements_dict_cache_file(self):
        return measurements.universal.constants.MEASUREMENTS_DICT_FILE.format(tracer=self.tracer, data_set=self.data_set_name)


    ## means

    def _mean_cache_file(self, target):
        fill_strategy=self._fill_strategy_id('concentration_means')
        return measurements.universal.constants.MEAN_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements=self.min_measurements_mean, fill_strategy=fill_strategy, target=target)


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_mean'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def means(self):
        return super().means

    def means_cache_file(self):
        return self._mean_cache_file('sample_points')


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_mean'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def means_for_sample_lsm(self):
        return super().means_for_sample_lsm

    def means_for_sample_lsm_cache_file(self):
        return self._mean_cache_file(str(self.sample_lsm))


    ## deviation

    def _standard_deviations_cache_file(self, deviation_type, target):
        fill_strategy = self._fill_strategy_id(deviation_type)
        if deviation_type == 'concentration_standard_deviations':
            min_standard_deviation = 0
        else:
            min_standard_deviation = self.min_standard_deviation
        return measurements.universal.constants.DEVIATION_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements=self.min_measurements_standard_deviation, min_standard_deviation=min_standard_deviation, deviation_type=deviation_type, fill_strategy=fill_strategy, target=target)


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def concentration_standard_deviations(self):
        return super().concentration_standard_deviations

    def concentration_standard_deviations_cache_file(self):
        return self._standard_deviations_cache_file('concentration_standard_deviations', 'sample_points')

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def concentration_standard_deviations_for_sample_lsm(self):
        return super().concentration_standard_deviations_for_sample_lsm

    def concentration_standard_deviations_for_sample_lsm_cache_file(self):
        return self._standard_deviations_cache_file('concentration_standard_deviations', str(self.sample_lsm))


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def average_noise_standard_deviations(self):
        return super().average_noise_standard_deviations

    def average_noise_standard_deviations_cache_file(self):
        return self._standard_deviations_cache_file('average_noise_standard_deviations', 'sample_points')

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def average_noise_standard_deviations_for_sample_lsm(self):
        return super().average_noise_standard_deviations_for_sample_lsm

    def average_noise_standard_deviations_for_sample_lsm_cache_file(self):
        return self._standard_deviations_cache_file('average_noise_standard_deviations', str(self.sample_lsm))


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def noise_standard_deviations(self):
        return super().noise_standard_deviations

    def noise_standard_deviations_cache_file(self):
        return self._standard_deviations_cache_file('noise_standard_deviations', 'sample_points')


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def standard_deviations(self):
        return super().standard_deviations

    def standard_deviations_cache_file(self):
        return self._standard_deviations_cache_file('standard_deviations', 'sample_points')

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def standard_deviations_for_sample_lsm(self):
        return super().standard_deviations_for_sample_lsm

    def standard_deviations_for_sample_lsm_cache_file(self):
        return self._standard_deviations_cache_file('standard_deviations', str(self.sample_lsm))


    ## correlation

    @property
    #@overrides.overrides
    def _sample_correlation(self):
        return measurements.universal.sample_data.SampleCorrelationMatrixCache(self, self.sample_lsm, self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, matrix_format=self.matrix_format_correlation, dtype=self.dtype_correlation)

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation', 'self.min_measurements_correlation', 'self.min_abs_correlation', 'self.max_abs_correlation', 'self.cholesky_min_diag_value_correlation', 'self.cholesky_ordering_method_correlation', 'self.cholesky_reordering_correlation', 'self.matrix_format_correlation', 'self.dtype_correlation'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def correlations_own(self):
        import util.math.sparse.decompose.with_cholmod
        correlation_matrix, reduction_factors = util.math.sparse.decompose.with_cholmod.approximate_positive_definite(self.correlations_own_sample_matrix, min_abs_value=self.min_abs_correlation, min_diag_value=self.cholesky_min_diag_value_correlation, ordering_method=self.cholesky_ordering_method_correlation, reorder_after_each_step=self.cholesky_reordering_correlation, reduction_factors_file=self.reduction_factors_cache_file())
        return correlation_matrix.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)

    def correlations_own_cache_file(self):
        return measurements.universal.constants.CORRELATION_MATRIX_POSITIVE_DEFINITE_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, cholesky_ordering_method_correlation=self.cholesky_ordering_method_correlation, cholesky_reordering_correlation=self.cholesky_reordering_correlation, cholesky_min_diag_value=self.cholesky_min_diag_value_correlation, standard_deviation_id=self.standard_deviation_id_without_sample_lsm, dtype=self.dtype_correlation, matrix_format=self.matrix_format_correlation)

    def reduction_factors_cache_file(self):
        return measurements.universal.constants.CORRELATION_MATRIX_POSITIVE_DEFINITE_REDUCTION_FACTORS_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, cholesky_ordering_method_correlation=self.cholesky_ordering_method_correlation, cholesky_reordering_correlation=self.cholesky_reordering_correlation, cholesky_min_diag_value=self.cholesky_min_diag_value_correlation, standard_deviation_id=self.standard_deviation_id_without_sample_lsm)


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation', 'self.min_measurements_correlation', 'self.min_abs_correlation', 'self.max_abs_correlation', 'self.cholesky_min_diag_value_correlation', 'self.cholesky_ordering_method_correlation', 'self.cholesky_reordering_correlation', 'self.matrix_format_correlation', 'self.dtype_correlation'))
    def _correlations_own_cholesky_decomposition(self):
        return super().correlations_own_cholesky_decomposition

    @property
    @util.cache.file.decorator()
    def _correlations_own_cholesky_decomposition_P(self):
        return self._correlations_own_cholesky_decomposition['P']

    def _correlations_own_cholesky_decomposition_P_cache_file(self):
        return measurements.universal.constants.CORRELATION_MATRIX_CHOLESKY_FACTOR_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, cholesky_ordering_method_correlation=self.cholesky_ordering_method_correlation, cholesky_reordering_correlation=self.cholesky_reordering_correlation, cholesky_min_diag_value=self.cholesky_min_diag_value_correlation, standard_deviation_id=self.standard_deviation_id_without_sample_lsm, dtype=np.dtype(np.int8), matrix_format=self.matrix_format_correlation, factor_type='P')

    @property
    @util.cache.file.decorator()
    def _correlations_own_cholesky_decomposition_L(self):
        return self._correlations_own_cholesky_decomposition['L']

    def _correlations_own_cholesky_decomposition_L_cache_file(self):
        return measurements.universal.constants.CORRELATION_MATRIX_CHOLESKY_FACTOR_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, min_measurements_correlation=self.min_measurements_correlation, min_abs_correlation=self.min_abs_correlation, max_abs_correlation=self.max_abs_correlation, cholesky_ordering_method_correlation=self.cholesky_ordering_method_correlation, cholesky_reordering_correlation=self.cholesky_reordering_correlation, cholesky_min_diag_value=self.cholesky_min_diag_value_correlation, standard_deviation_id=self.standard_deviation_id_without_sample_lsm, dtype=self.dtype_correlation, matrix_format=self.matrix_format_correlation, factor_type='L')


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name', 'self.fill_strategy', 'self.min_measurements_standard_deviation', 'self.min_standard_deviation', 'self.min_measurements_correlation', 'self.min_abs_correlation', 'self.max_abs_correlation', 'self.cholesky_min_diag_value_correlation', 'self.cholesky_ordering_method_correlation', 'self.cholesky_reordering_correlation', 'self.matrix_format_correlation', 'self.dtype_correlation'))
    #@overrides.overrides
    def correlations_own_cholesky_decomposition(self):
        return {'P': self._correlations_own_cholesky_decomposition_P, 'L': self._correlations_own_cholesky_decomposition_L}




class MeasurementsAnnualPeriodicNearWaterCache(MeasurementsAnnualPeriodicCache, MeasurementsAnnualPeriodicNearWater):

    ## ids

    @property
    #@overrides.overrides
    def mean_id(self):
        return self.base_measurements.mean_id

    @property
    #@overrides.overrides
    def standard_deviation_id(self):
        return self.base_measurements.standard_deviation_id

    @property
    #@overrides.overrides
    def correlation_id(self):
        return self.base_measurements.correlation_id


    ## cacheable properties

    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def near_water_projection_matrix(self):
        return super().near_water_projection_matrix


    @property
    @util.cache.memory.method_decorator(dependency=('self.tracer', 'self.data_set_name'))
    @util.cache.file.decorator()
    #@overrides.overrides
    def correlations_own_sample_matrix(self):
        return super().correlations_own_sample_matrix


    ## cache files

    def near_water_projection_matrix_cache_file(self):
        return measurements.universal.constants.NEAR_WATER_PROJECTION_MASK_FILE.format(tracer=self.tracer, data_set=self.data_set_name, sample_lsm=self.sample_lsm, water_lsm=self.water_lsm, max_box_distance_to_water=self.max_box_distance_to_water, matrix_format='csc')

    #@overrides.overrides
    def points_cache_file(self):
        return self.base_measurements.points_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def values_cache_file(self):
        return self.base_measurements.values_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def measurements_dict_cache_file(self):
        return self.base_measurements.measurements_dict_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def means_cache_file(self):
        return self.base_measurements.means_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def concentration_standard_deviations_cache_file(self):
        return self.base_measurements.concentration_standard_deviations_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def noise_standard_deviations_cache_file(self):
        return self.base_measurements.noise_standard_deviations_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def average_noise_standard_deviations_cache_file(self):
        return self.base_measurements.average_noise_standard_deviations_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def standard_deviations_cache_file(self):
        return self.base_measurements.standard_deviations_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    def correlations_own_sample_matrix_cache_file(self):
        return self.base_measurements._sample_correlation.correlation_matrix_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def correlations_own_cache_file(self):
        return self.base_measurements.correlations_own_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def reduction_factors_cache_file(self):
        return self.base_measurements.reduction_factors_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def _correlations_own_cholesky_decomposition_P_cache_file(self):
        return self.base_measurements._correlations_own_cholesky_decomposition_P_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)

    #@overrides.overrides
    def _correlations_own_cholesky_decomposition_L_cache_file(self):
        return self.base_measurements._correlations_own_cholesky_decomposition_L_cache_file().replace(self.base_measurements.data_set_name, self.data_set_name)




class MeasurementsAnnualPeriodicUnionCache(MeasurementsAnnualPeriodicUnion, MeasurementsAnnualPeriodicCache):
    pass




class MeasurementsCollectionCache(MeasurementsCache, MeasurementsCollection):

    ## ids

    @property
    #@overrides.overrides
    def mean_id(self):
        return util.str.merge([measurement.mean_id for measurement in self.measurements_list])

    @property
    #@overrides.overrides
    def standard_deviation_id(self):
        return util.str.merge([measurement.standard_deviation_id for measurement in self.measurements_list])

    @property
    #@overrides.overrides
    def correlation_id(self):
        return util.str.merge([measurement.correlation_id for measurement in self.measurements_list])


    ## cached values

    @property
    @util.cache.file.decorator()
    #@overrides.overrides
    def correlations_own_sample_matrix(self):
        return super().correlations_own_sample_matrix

    @property
    @util.cache.file.decorator()
    #@overrides.overrides
    def correlations_own(self):
        import util.math.sparse.decompose.with_cholmod
        correlation_matrix, reduction_factors = util.math.sparse.decompose.with_cholmod.approximate_positive_definite(self.correlations_own_sample_matrix, min_abs_value=self.min_abs_correlation, min_diag_value=self.cholesky_min_diag_value_correlation, ordering_method=self.cholesky_ordering_method_correlation, reorder_after_each_step=self.cholesky_reordering_correlation, reduction_factors_file=self.reduction_factors_cache_file())
        return correlation_matrix.asformat(self.matrix_format_correlation).astype(self.dtype_correlation)

    @property
    @util.cache.memory.method_decorator()
    def _correlations_own_cholesky_decomposition(self):
        return super().correlations_own_cholesky_decomposition

    @property
    @util.cache.file.decorator()
    def _correlations_own_cholesky_decomposition_P(self):
        return self._correlations_own_cholesky_decomposition['P']

    @property
    @util.cache.file.decorator()
    def _correlations_own_cholesky_decomposition_L(self):
        return self._correlations_own_cholesky_decomposition['L']

    @property
    #@overrides.overrides
    def correlations_own_cholesky_decomposition(self):
        return {'P': self._correlations_own_cholesky_decomposition_P, 'L': self._correlations_own_cholesky_decomposition_L}


    ## files

    def _merge_files(self, directory, files):
        ## common dirnames above file
        number_of_measurement_dirs_below_base_dir = measurements.universal.constants.MEASUREMENT_DIR[len(measurements.universal.constants.BASE_DIR):].count(os.path.sep)
        filenames = [file[len(measurements.universal.constants.BASE_DIR):] for file in files]
        filenames = [os.path.join(*file.split(os.path.sep)[number_of_measurement_dirs_below_base_dir+1:]) for file in filenames]

        ## join dirs and filename
        filename_joined = util.str.merge(filenames)
        file_joined = os.path.join(directory, filename_joined)
        return file_joined

    @property
    def measurements_dir(self):
        return measurements.universal.constants.MEASUREMENT_DIR.format(tracer=self.tracer, data_set=self.data_set_name)

    def correlations_own_sample_matrix_cache_file(self):
        merged_correlation_ids = util.str.merge([measurement.correlation_id for measurement in self.measurements_list])
        filename = measurements.universal.constants.SEPERATOR.join(['sample_correlation', merged_correlation_ids, '{dtype}.{matrix_format}.npz'.format(dtype=self.dtype_correlation, matrix_format=self.matrix_format_correlation)])
        return os.path.join(self.measurements_dir, 'correlation', 'sample_correlation', filename)

    def reduction_factors_cache_file(self):
        return self._merge_files(self.measurements_dir, [measurement.reduction_factors_cache_file() for measurement in self.measurements_list])

    def correlations_own_cache_file(self):
        return self._merge_files(self.measurements_dir, [measurement.correlations_own_cache_file() for measurement in self.measurements_list])

    def _correlations_own_cholesky_decomposition_P_cache_file(self):
        return self._merge_files(self.measurements_dir, [measurement._correlations_own_cholesky_decomposition_P_cache_file() for measurement in self.measurements_list])

    def _correlations_own_cholesky_decomposition_L_cache_file(self):
        return self._merge_files(self.measurements_dir, [measurement._correlations_own_cholesky_decomposition_L_cache_file() for measurement in self.measurements_list])




## generic

class TooFewValuesError(Exception):

    def __init__(self):
        message = 'Too few values are available.'
        super().__init__(message)




def as_measurements_collection(measurements):
    if isinstance(measurements, MeasurementsCollectionCache):
        return measurements
    else:
        return MeasurementsCollectionCache(*measurements)




import util.io.env

BASE_DIR_ENV_NAME = 'MEASUREMENTS_DIR'
BASE_DIR = util.io.env.load(BASE_DIR_ENV_NAME)

## earth
EARTH_RADIUS = 6371 * 10**3
MAX_SEA_DEPTH = 11 * 10**3

## mean
MEAN_MIN_MEASUREMENTS = 2

## deviation
DEVIATION_MIN_MEASUREMENTS = 3
DEVIATION_MIN_VALUE = 0.05

## correlation
CORRELATION_MIN_ABS_VALUE = 0.01
CORRELATION_MAX_ABS_VALUE = 0.99
CORRELATION_MIN_DIAG_VALUE_POSITIVE_DEFINITE_APPROXIMATION = 0.1



import abc
import bisect

import numpy as np

import util.math.interpolate
import util.math.spherical

import util.logging
logger = util.logging.logger



class Time_Periodic_Earth_Interpolator(util.math.interpolate.Periodic_Interpolator):

    def __init__(self, data_points, data_values, t_len, wrap_around_amount=0, number_of_linear_interpolators=1, single_overlapping_amount_linear_interpolators=0, parallel=False):
        from measurements.constants import EARTH_RADIUS

        logger.debug('Initiating time periodic earth interpolator with {} data points, time len {}, wrap around amount {} and {} linear interpolators with single overlapping amount of {}.'.format(len(data_points), t_len, wrap_around_amount, number_of_linear_interpolators, single_overlapping_amount_linear_interpolators))

        ## call super constructor
        self.order = number_of_linear_interpolators

        t_scaling = 2 * EARTH_RADIUS / t_len

        super().__init__(data_points, data_values, point_range_size=(t_len, None, None, None), wrap_around_amount=(wrap_around_amount, 0, 0, 0),  scaling_values=(t_scaling, None, None, None), number_of_linear_interpolators=number_of_linear_interpolators, single_overlapping_amount_linear_interpolators=single_overlapping_amount_linear_interpolators, parallel=parallel)

        assert len(self._data_points) == len(self._data_values) == len(self._data_indices)


    def _modify_points(self, points, is_data_points):
        from measurements.constants import EARTH_RADIUS, MAX_SEA_DEPTH

        points = super()._modify_points(points, is_data_points)

        ## if data points, append values for lower and upper bound of depth
        if self.order > 0 and is_data_points:
            lower_depth = 0
            lower_depth_bound = np.min(points[:,3])
            upper_depth = MAX_SEA_DEPTH
            upper_depth_bound = np.max(points[:,3])

            logger.debug('Lower depth is {}, upper depth is {}.'.format(lower_depth, upper_depth))

            assert lower_depth_bound >= lower_depth and upper_depth_bound <= upper_depth

            if lower_depth_bound > lower_depth:
                lower_depth_bound_indices = np.where(np.isclose(points[:,3], lower_depth_bound))[0]
                lower_depth_bound_points = points[lower_depth_bound_indices]
                lower_depth_bound_points[:,3] = lower_depth
                logger.debug('{} values appended for lower bound {}.'.format(len(lower_depth_bound_indices), lower_depth))
            else:
                lower_depth_bound_indices = np.array([])
                lower_depth_bound_points = np.array([])
                logger.debug('No values appended for lower bound {}.'.format(lower_depth))
            if upper_depth_bound < upper_depth:
                upper_depth_bound_indices = np.where(np.isclose(points[:,3], lower_depth_bound))[0]
                upper_depth_bound_points = points[upper_depth_bound_indices]
                upper_depth_bound_points[:,3] = upper_depth
                logger.debug('{} values appended for upper bound {}.'.format(len(upper_depth_bound_indices), upper_depth))
            else:
                upper_depth_bound_indices= np.array([])
                upper_depth_bound_points = np.array([])
                logger.debug('No values appended for upper bound {}.'.format(upper_depth))

            indices = np.concatenate((lower_depth_bound_indices, np.arange(len(points)), upper_depth_bound_indices), axis=0)
            points = np.concatenate((lower_depth_bound_points, points, upper_depth_bound_points), axis=0)
            self._data_indices = self._data_indices[indices]

        ## convert to cartesian
        points[:,1:] =  util.math.spherical.to_cartesian(points[:,1:], surface_radius=EARTH_RADIUS)

        return points



def periodic_with_coordinates(data, interpolation_points, lsm_base, scaling_values=None, interpolator_options=None):
    logger.debug('Interpolating periodic data with coordinates for lsm {} with scaling_values {} and interpolator_options {}.'.format(lsm_base, scaling_values, interpolator_options))
    
    ## convert coordinates to map indices
    data = np.array(data, copy=True)
    data[:, :-1] = lsm_base.coordinates_to_map_indices(data[:, :-1], discard_year=True, int_indices=False)
    interpolation_points = lsm_base.coordinates_to_map_indices(interpolation_points, discard_year=True, int_indices=False)

    ## interpolating
    return periodic_with_map_indices(data, interpolation_points, lsm_base, scaling_values=scaling_values, interpolator_options=interpolator_options)



def periodic_with_map_indices(data, interpolation_points, lsm_base, scaling_values=None, interpolator_options=None):
    logger.debug('Interpolating periodic data with map indices for lsm {} with scaling_values {} and interpolator_options {}.'.format(lsm_base, scaling_values, interpolator_options))
    
    assert data.ndim == 2
    assert data.shape[1] == 5
    assert interpolation_points.ndim == 2
    assert interpolation_points.shape[1] == 4

    ## split in points and values
    data_points = data[:, :-1]
    data_values = data[:, -1]

    ## scaling values
    if scaling_values is None:
        scaling_values=(lsm_base.x_dim/lsm_base.t_dim, None, None, None)
    
    if interpolator_options is None:
        interpolator_options = (1,1,0,0)

    ## prepare wrap_around_amount
    wrap_around_amount = interpolator_options[0]
    try:
        wrap_around_amount = tuple(wrap_around_amount)
    except TypeError:
        wrap_around_amount = (wrap_around_amount,)
    if len(wrap_around_amount) == 1:
        ## use same wrap around for t and x
        wrap_around_amount = wrap_around_amount * 2
    if len(wrap_around_amount) == 2:
        ## append wrap around for y and z if missing
        wrap_around_amount = wrap_around_amount + (0,0)

    ## create interpolator
    interpolator = util.math.interpolate.Periodic_Interpolator(data_points, data_values, point_range_size=(lsm_base.t_dim, lsm_base.x_dim, lsm_base.y_dim, lsm_base.z_dim), scaling_values=scaling_values, wrap_around_amount=wrap_around_amount, number_of_linear_interpolators=interpolator_options[1], single_overlapping_amount_linear_interpolators=interpolator_options[2], parallel=bool(interpolator_options[3]))

    ## interpolating
    interpolation_data = interpolator.interpolate(interpolation_points)
    return interpolation_data



def default_scaling_values(sample_lsm):
    scaling_x = 1
    scaling_y = sample_lsm.x_dim / (sample_lsm.y_dim * 2)
    if scaling_y.is_integer():
        scaling_y = int(scaling_y)
    scaling_t = sample_lsm.x_dim / (sample_lsm.t_dim * 3)
    if scaling_t.is_integer():
        scaling_t = int(scaling_t)
    scaling_z = int(np.floor(sample_lsm.x_dim / sample_lsm.z_dim))
    scaling_values = (scaling_t, scaling_x, scaling_y, scaling_z)
    return scaling_values
    
    

class Interpolator_Annual_Periodic:

    def __init__(self, sample_lsm, scaling_values=None):
        self.sample_lsm = sample_lsm
        self.scaling_values = scaling_values


    @property
    def scaling_values(self):
        return self._scaling_values

    @scaling_values.setter
    def scaling_values(self, scaling_values):
        if scaling_values is None:
            scaling_values = default_scaling_values(self.sample_lsm)
        self._scaling_values = scaling_values
    

    def interpolate_data_for_lsm(self, data, lsm, interpolator_options=None):
        logger.debug('Interpolating data for lsm {} with interpolator_options {}.'.format(lsm, interpolator_options))
        
        sea_indices = lsm.sea_indices
        sea_coordinates = lsm.map_indices_to_coordinates(sea_indices)
        
        interpolated_values = periodic_with_coordinates(data, sea_coordinates, self.sample_lsm, scaling_values=self.scaling_values, interpolator_options=interpolator_options)
        interpolated_data = np.concatenate((sea_indices, interpolated_values[:,np.newaxis]), axis=1)
        interpolated_map = lsm.insert_index_values_in_map(interpolated_data, no_data_value=np.inf)
        assert np.all(interpolated_map != np.inf)
        
        util.math.interpolate.change_dim(interpolated_map, 0, lsm.t_dim)
        assert interpolated_map.shape == lsm.dim
        
        return interpolated_map
    

    def interpolate_data_for_sample_lsm_with_coordinates(self, data, interpolator_options=None):
        logger.debug('Interpolating data with coordinates for lsm {} with interpolator_options {}.'.format(self.sample_lsm, interpolator_options))
        data = np.array(data, copy=True)
        data[:, :-1] = lsm_base.coordinates_to_map_indices(data[:, :-1], discard_year=True, int_indices=False)
        return self.interpolate_data_for_sample_lsm_with_map_indices(data, self.sample_lsm, interpolator_options=interpolator_options)


    def interpolate_data_for_sample_lsm_with_map_indices(self, data, interpolator_options=None):
        logger.debug('Interpolating data with map indices for lsm {} with interpolator_options {}.'.format(self.sample_lsm, interpolator_options))
        
        sea_indices = self.sample_lsm.sea_indices
        
        interpolated_values = periodic_with_map_indices(data, sea_indices, self.sample_lsm, scaling_values=self.scaling_values, interpolator_options=interpolator_options)
        interpolated_data = np.concatenate((sea_indices, interpolated_values[:,np.newaxis]), axis=1)
        
        interpolated_map = self.sample_lsm.insert_index_values_in_map(interpolated_data, no_data_value=np.inf)
        assert np.all(interpolated_map != np.inf)
        util.math.interpolate.change_dim(interpolated_map, 0, self.sample_lsm.t_dim)
        assert interpolated_map.shape == self.sample_lsm.dim
        
        return interpolated_map


    def interpolate_data_for_points_from_interpolated_lsm_data(self, interpolated_lsm_data, interpolation_points):
        logger.debug('Interpolationg data for points from interpolated data for lsm {}.'.format(self.sample_lsm))
        
        ## get interpolated points and values
        interpolated_lsm_data_mask = ~np.isnan(interpolated_lsm_data)
        interpolated_lsm_values = interpolated_lsm_data[interpolated_lsm_data_mask]
        interpolated_lsm_indices = np.array(np.where(interpolated_lsm_data_mask)).T
        interpolated_lsm_points = self.sample_lsm.map_indices_to_coordinates(interpolated_lsm_indices)
        interpolated_lsm_points_and_values = np.concatenate([interpolated_lsm_points, interpolated_lsm_values[:,np.newaxis]], axis=1)
        
        ## prepare interpolation points: convert to (rounded) int map indices -> convert to coordinates
        interpolation_points_map_indices = self.sample_lsm.coordinates_to_map_indices(interpolation_points, discard_year=True, int_indices=True)
        interpolation_points = self.sample_lsm.map_indices_to_coordinates(interpolation_points_map_indices)

        ## interpolate for points
        interpolated_points_data = periodic_with_coordinates(interpolated_lsm_points_and_values, interpolation_points, self.sample_lsm, scaling_values=self.scaling_values, interpolator_options=(2/min([self.sample_lsm.t_dim,self.sample_lsm.x_dim]),0,0,0))
        
        ## return
        assert np.all(np.isfinite(interpolated_points_data))
        return interpolated_points_data



    def interpolate_data_for_points(self, data, interpolation_points, interpolator_options=None):
        logger.debug('Interpolationg data for points with interpolator_options {}.'.format(interpolator_options))
        
        ## interpolate for sample lsm
        interpolated_lsm_data = self.interpolate_data_for_lsm(data, self.sample_lsm, interpolator_options=interpolator_options)
        
        ## interpolate for points
        interpolated_points_data = self.interpolate_data_for_points_from_interpolated_lsm_data(interpolated_lsm_data, interpolation_points)
        return interpolated_points_data
        
        # ## get interpolated points and values
        # interpolated_lsm_data_mask = ~np.isnan(interpolated_lsm_data)
        # interpolated_lsm_values = interpolated_lsm_data[interpolated_lsm_data_mask]
        # interpolated_lsm_indices = np.array(np.where(interpolated_lsm_data_mask)).T
        # interpolated_lsm_points = self.sample_lsm.map_indices_to_coordinates(interpolated_lsm_indices)
        # interpolated_lsm_points_and_values = np.concatenate([interpolated_lsm_points, interpolated_lsm_values[:,np.newaxis]], axis=1)
        # 
        # ## prepare interpolation points: convert to (rounded) int map indices -> convert to coordinates
        # interpolation_points_map_indices = self.sample_lsm.coordinates_to_map_indices(interpolation_points, discard_year=True, int_indices=True)
        # interpolation_points = self.sample_lsm.map_indices_to_coordinates(interpolation_points_map_indices)

        #   ## interpolate for points
        # interpolated_data = periodic_with_coordinates(interpolated_lsm_points_and_values, interpolation_points, self.sample_lsm, scaling_values=self.scaling_values, interpolator_options=(2/min([self.sample_lsm.t_dim,self.sample_lsm.x_dim]),0,0,0))
        # 
        # ## return
        # assert np.all(np.isfinite(interpolated_data))
        # return interpolated_data




import datetime
import os.path

import measurements.land_sea_mask.lsm
from measurements.constants import BASE_DIR


## informations for the wod cruises

BASE_DATE = datetime.datetime(1770, 1, 1)
DAY_OFFSET = 'time' # number of days since 01.01.1770 (float)
DAY_OFFSET_UNIT = b'days since 1770-01-01 00:00:00'

LAT = 'lat'
LAT_UNIT = b'degrees_north'
LON = 'lon'
LON_UNIT = b'degrees_east'
DEPTH = 'z'
DEPTH_UNIT = b'm'

PO4 = 'Phosphate'
PO4_UNIT = b'umol/l'

DEPTH_FLAG = 'z_WODflag'
PO4_FLAG = 'Phosphate_WODflag'
PO4_PROFILE_FLAG = 'Phosphate_WODprofileflag'
MISSING_VALUE = - 10**10

## data dirs and files

WOD_DIR = os.path.join(BASE_DIR, 'po4', 'wod_2013')
DATA_DIR = os.path.join(WOD_DIR, 'data')
CRUISES_DATA_DIR = os.path.join(DATA_DIR, 'cruises')
CRUISES_LIST_FILE = os.path.join(DATA_DIR, 'cruises_list.ppy')
POINTS_AND_RESULTS_FILE = os.path.join(DATA_DIR, 'points_and_results.npz')
MEASUREMENTS_DICT_SORTED_FILE = os.path.join(DATA_DIR, 'measurement_dict_sorted.ppy')
MEASUREMENTS_DICT_UNSORTED_FILE = os.path.join(DATA_DIR, 'measurement_dict_unsorted.ppy')


## sample lsm

SAMPLE_T_DIM = 12
SAMPLE_LSM = measurements.land_sea_mask.lsm.LandSeaMaskWOA13R(t_dim=SAMPLE_T_DIM)



from util.logging import Logger
from interface import plot_correlogram as plot

with Logger():
    for min_measurements in (10, 25, 50, 100, 200, 500):
        plot(show_model=False, min_measurements=min_measurements)


import bisect
import logging

import numpy as np

import util.math.interpolate


## convert point to box index (helper functions)

def get_spatial_float_index(x, y, z, land_sea_mask, z_values_left, x_range=[0, 360], y_range=[-90, 90]):
    logging.debug('Getting spatial float index for {}.'.format((x, y, z)))

    ## adjust x coordinates if negative
    if x < 0:
        x += 360

    ## check input
    if x < x_range[0] or x > x_range[1]:
        raise ValueError('Value {} of x is not in range {}.'.format(x, x_range))
    if y < y_range[0] or y > y_range[1]:
        raise ValueError('Value {} of y is not in range {}.'.format(y, y_range))
    if z < z_values_left[0]:
        raise ValueError('Value {} of z have to be greater or equal to {}.'.format(z, z_values_left[0]))

    ## linear interpolate x and y index
    (x_dim, y_dim) = land_sea_mask.shape
    x_index_float = util.math.interpolate.get_float_index_for_equidistant_values(x, x_range, x_dim)
    y_index_float = util.math.interpolate.get_float_index_for_equidistant_values(y, y_range, y_dim)

    ## lockup z
    z_index = bisect.bisect_right(z_values_left, z) - 1

    if z_index + 1 < len(z_values_left):
        z_left = z_values_left[z_index]
        z_right = z_values_left[z_index + 1]
        z_index_float = z_index + (z - z_left) / (z_right - z_left)
    else:
        z_index_float = z_index


    logging.debug('Float indices for {} are {}.'.format((x, y, z), (x_index_float, y_index_float, z_index_float)))

    return (x_index_float, y_index_float, z_index_float)


def get_temporal_float_index(t, t_dim, t_range=[0, 1]):
    logging.debug('Getting temporal float index for {}.'.format(t))

    ## check input
    if t < t_range[0] or t > t_range[1]:
        raise ValueError('Value {} of t is not in range {}.'.format(t, t_range))

    ## interpolate
    t_index_float = util.math.interpolate.get_float_index_for_equidistant_values(t, t_range, t_dim)

    logging.debug('Temporal float index for {} is {}.'.format(t, t_index_float))

    return t_index_float


def is_water(spatial_indices, land_sea_mask):
    logging.debug('Checking if indices are water.')

    ## check input
    spatial_indices = np.array(spatial_indices, dtype=np.int)

    dim = len(spatial_indices.shape)

    if dim > 2 or dim == 0:
        raise ValueError('The spatial_indices array has to have 1 or 2 dimensions, but it has {} dimensions.'.format(dim))
    elif dim == 1:
        spatial_len = spatial_indices.shape[0]
        if spatial_len != 3:
            raise ValueError('The len of spatial_indices array has to be 3, but it is {}.'.format(spatial_len))
        spatial_indices = spatial_indices.reshape([1, spatial_len])
    else:
        spatial_len = spatial_indices.shape[1]
        if spatial_len != 3:
            raise ValueError('The second dimension of the spatial_indices array to be 3, but it is {}.'.format(spatial_len))

    ## compute if water
    land_sea_mask_indices = spatial_indices[:, :-1]
    land_sea_mask_values = land_sea_mask[land_sea_mask_indices[:, 0], land_sea_mask_indices[:, 1]]
    is_water = np.logical_or(np.isnan(land_sea_mask_values),  land_sea_mask_values >= spatial_indices[:, -1])

    return is_water


def get_all_water_boxes(land_sea_mask):
    logging.debug('Getting all water boxes.')

    land_sea_mask = land_sea_mask.lsm
    (water_x, water_y) = np.where(land_sea_mask != 0)
    water_len = np.sum(land_sea_mask)
    water_boxes = np.empty([water_len, 3], dtype=np.int)

    j = 0
    for i in range(len(water_x)):
        x_i = water_x[i]
        y_i = water_y[i]
        z_i = land_sea_mask[x_i, y_i]

        j_range = range(j, j + z_i)
        water_boxes[j_range, 0] = x_i
        water_boxes[j_range, 1] = y_i
        water_boxes[j_range, 2] = range(z_i)

        j += z_i

    return water_boxes


def get_nearest_water_box(land_sea_mask, x_index, y_index, z_index):
    logging.debug('Getting nearest water box for index {}.'.format((x_index, y_index, z_index)))

    water_boxes = get_all_water_boxes(land_sea_mask)

    index = [x_index, y_index, z_index]
    nearest_water_box = util.math.interpolate.get_nearest_value_in_array(water_boxes, index)

    assert land_sea_mask[nearest_water_box[0], nearest_water_box[1]] >= nearest_water_box[2]

    return nearest_water_box


def get_nearest_spatial_water_index(x, y, z, land_sea_mask, z_values_left, x_range=[0, 360], y_range=[-90, 90]):
    (x_index_float, y_index_float, z_index_float) = get_spatial_float_index(x, y, z, land_sea_mask, z_values_left, x_range=x_range, y_range=y_range)

    ## floor indices to int
    x_index = np.floor(x_index_float)
    y_index = np.floor(y_index_float)
    z_index = np.floor(z_index_float)

    ## get nearest water box if box is land
    if not is_water((x_index, y_index, z_index), land_sea_mask):
        logging.debug('Box {} is land.'.format((x_index, y_index, z_index)))
        (x_index, y_index, z_index) = get_nearest_water_box(land_sea_mask, x_index_float, y_index_float, z_index_float)

    logging.debug('Nearest index for {} is {}.'.format((x, y, z), (x_index_float, y_index_float, z_index_float)))

    assert land_sea_mask[x_index, y_index] >= z_index

    return (x_index, y_index, z_index)


def get_nearest_temporal_index(t, t_dim, t_range=[0, 1]):
    t_index_float = get_temporal_float_index(t, t_dim, t_range=t_range)
    t_index = np.floor(t_index_float)

    logging.debug('Nearest temporal index for {} is {}.'.format(t, t_index))

    return t_index


def get_nearest_water_index(t, x, y, z, land_sea_mask, z_values_left, t_range=[0, 1], x_range=[0, 360], y_range=[-90, 90]):
    t_index =  get_nearest_temporal_index(t, land_sea_mask.t_dim, t_range=t_range)
    x_index, y_index, z_index = get_nearest_spatial_water_index(x, y, z, land_sea_mask, z_values_left, x_range=x_range, y_range=y_range)

    assert land_sea_mask[x_index, y_index] >= z_index

    return (t_index, x_index, y_index, z_index)


def convert_point_to_box_index(t, x, y, z, land_sea_mask, z_values_left, t_range=[0, 1], x_range=[0, 360], y_range=[-90, 90]):
    return get_nearest_water_index(t, x, y, z, land_sea_mask, z_values_left, t_range=t_range, x_range=x_range, y_range=y_range)




## regird

def measurements_to_land_sea_mask(measurement_data, land_sea_mask, t_range=[0, 1], x_range=[0, 360], y_range=[-90, 90]):
    assert measurement_data.ndim == 2
    assert measurement_data.shape[1] in [5, 7]

    z_values_left = land_sea_mask.z_left

    ## init arrays
    nobs = land_sea_mask.masked_map(default_value=0, dtype=np.float64)
    sum_of_values = np.copy(nobs)
    sum_of_squares = np.copy(nobs)
    variances = land_sea_mask.masked_map(default_value=np.inf, dtype=np.float64)

    ## init dim values
    number_of_measurements = measurement_data.shape[0]
    data_dim = measurement_data.shape[1]

    ## insert measurements
    for i in range(number_of_measurements):
        t, x, y, z, data_sum_of_values = measurement_data[i, :5]
        if data_dim == 5:
            data_nobs = 1
            data_sum_of_squares = data_sum_of_values**2
        else:
            data_nobs, data_sum_of_squares = measurement_data[i, 5:]

        (t_index, x_index, y_index, z_index) = convert_point_to_box_index(t, x, y, z, land_sea_mask, z_values_left, t_range=t_range, x_range=x_range, y_range=y_range)

        nobs[t_index, x_index, y_index, z_index] += data_nobs
        sum_of_values[t_index, x_index, y_index, z_index] += data_sum_of_values
        sum_of_squares[t_index, x_index, y_index, z_index] += data_sum_of_squares

    ## calculate mean
    where_measurements = np.where(nobs > 0)

    means = np.copy(sum_of_values)
    means[where_measurements] /= nobs[where_measurements]

    ## calculate variance
    where_measurements_over_threshold = np.where(nobs >= 2)

    nobs_over_threshold = nobs[where_measurements_over_threshold]
    sum_of_values_over_threshold = sum_of_values[where_measurements_over_threshold]
    sum_of_squares_over_threshold = sum_of_squares[where_measurements_over_threshold]

    variances[where_measurements_over_threshold] = (sum_of_squares_over_threshold - sum_of_values_over_threshold ** 2 / nobs_over_threshold) / (nobs_over_threshold - 1)

    return (means, nobs, variances)



import os.path

from measurements.constants import BASE_DIR


## DOP
WOA_BASE_DIR = os.path.join(BASE_DIR, 'dop/woa')

NOBS_FILE =  os.path.join(WOA_BASE_DIR, 'dop_2.8x2.8_monthly_nobs.npy')
VARIS_FILE = os.path.join(WOA_BASE_DIR, 'dop_2.8x2.8_monthly_vari.npy')
MEANS_FILE = os.path.join(WOA_BASE_DIR, 'dop_2.8x2.8_monthly_mean.npy')


from measurements.constants import DEVIATION_MIN_MEASUREMENTS, DEVIATION_MIN_VALUE, MEAN_MIN_MEASUREMENTS, CORRELATION_MIN_ABS_VALUE, CORRELATION_MAX_ABS_VALUE

T_RANGE = [0, 1]
X_RANGE = [0, 360]
Y_RANGE = [-90, 90]


import datetime

import numpy as np
import scipy.io

import measurements.po4.wod.constants

import util.io.fs
import util.io.object
import util.datetime

import util.logging
logger = util.logging.logger


class UnitError(ValueError):
    def __init__(self, name, unit, wanted_unit):
        message = 'The variable {} has unit {} but unit {} was needed!'.format(name, unit, wanted_unit)
        super.__init__(message)

class MissingError(KeyError):
    def __init__(self, name):
        message = 'The variable {} is missing!'.format(name)
        super.__init__(message)



class Cruise():

    def __init__(self, file):

        def value_in_file(file, name, unit=None):
            try:
                var = file.variables[name]
            except KeyError as e:
                raise MissingError(name) from e
            else:
                if unit is not None and var.unit != unit:
                    raise UnitError(name, var.unit, unit)
                return var.data

        ## open netcdf file
        with scipy.io.netcdf.netcdf_file(file, 'r') as file:

            ## read time and data
            try:
                day_offset = value_in_file(file, measurements.po4.wod.constants.DAY_OFFSET, unit=measurements.po4.wod.constants.DAY_OFFSET_UNIT)
            except MissingError:
                time = float('nan')
            else:
                day_offset = float(day_offset)
                hours_offset = (day_offset % 1) * 24
                minutes_offset = (hours_offset % 1) * 60
                seconds_offset = (minutes_offset % 1) * 60

                day_offset = int(day_offset)
                hours_offset = int(hours_offset)
                minutes_offset = int(minutes_offset)
                seconds_offset = int(seconds_offset)

                dt_offset = datetime.timedelta(days=day_offset, hours=hours_offset, minutes=minutes_offset, seconds=seconds_offset)
                dt = measurements.po4.wod.constants.BASE_DATE + dt_offset
                time = util.datetime.datetime_to_float(dt)

            ## read coordinates and depth
            try:
                lon = value_in_file(file, measurements.po4.wod.constants.LON, unit=measurements.po4.wod.constants.LON_UNIT)
            except MissingError:
                lon = float('nan')
            else:
                lon = float(lon)
                if lon == 180:
                    lon = -180
                assert lon >= -180 and lon < 180

            try:
                lat = value_in_file(file, measurements.po4.wod.constants.LAT, unit=measurements.po4.wod.constants.LAT_UNIT)
            except MissingError:
                lat = float('nan')
            else:
                lat = float(lat)
                if lat == -90 or lat == 90:
                    lon = 0
                assert lat >= -90 and lat <= 90

            try:
                depth = value_in_file(file, measurements.po4.wod.constants.DEPTH, unit=measurements.po4.wod.constants.DEPTH_UNIT)
            except MissingError:
                depth = np.array([])

            ## read value
            try:
                values = value_in_file(file, measurements.po4.wod.constants.PO4, unit=measurements.po4.wod.constants.PO4_UNIT)
            except MissingError:
                values = np.array([])

            ## remove invalid measurements
            if len(values) > 0:
                z_flag = value_in_file(file, measurements.po4.wod.constants.DEPTH_FLAG)
                po4_flag = value_in_file(file, measurements.po4.wod.constants.PO4_FLAG)
                po4_profile_flag = value_in_file(file, measurements.po4.wod.constants.PO4_PROFILE_FLAG)

                valid_mask = np.logical_and(po4_flag == 0, z_flag == 0) * (po4_profile_flag == 0)
                valid_mask = np.logical_and(valid_mask, values != measurements.po4.wod.constants.MISSING_VALUE)
                depth = depth[valid_mask]
                values = values[valid_mask]


            ## check values
            if np.any(values < 0):
                logger.warn('Values in {} are lower then 0!'.format(file))
                valid_mask = values > 0
                values = values[valid_mask]
                depth = depth[valid_mask]

            if np.any(depth < 0):
                logger.warn('Depth in {} is lower then 0!'.format(file))
                depth[depth < 0] = 0


        ## save values
        self.time = time
        self.lon = lon
        self.lat = lat
        self.depth = depth
        self.values = values

        logger.debug('Cruise from {} loaded with {:d} values.'.format(file, self.number_of_measurements))



class CruiseCollection():

    def __init__(self, cruises=None):
        self.__cruises = cruises


    @property
    def cruises(self):
        try:
            cruises = self.__cruises
        except AttributeError:
            cruises = None

        return cruises

    @cruises.setter
    def cruises(self, cruises):
        self.__cruises = cruises


    def load_cruises_from_netcdf_files(self, data_dir):
        logger.debug('Loading all cruises from netcdf files.')

        ## lookup files
        logger.debug('Looking up files in %s.' % data_dir)
        files = util.io.fs.get_files(data_dir, use_absolute_filenames=True)
        logger.debug('%d files found.' % len(files))

        ## load cruises
        logger.debug('Loading cruises from found files.')
        cruises = [Cruise(file) for file in files]
        logger.debug('%d cruises loaded.' % len(cruises))

        ## remove empty cruises
        logger.debug('Removing empty cruises.')
        cruises = [cruise for cruise in cruises if cruise.number_of_measurements > 0]
        logger.debug('%d not empty cruises found.' % len(cruises))

        ## return cruises
        self.cruises = cruises


    def save_cruises_to_pickle_file(self, file):
        logger.debug('Saving cruises at %s.' % file)
        util.io.object.save(file, self.cruises)
        logger.debug('Cruises saved at %s.' % file)


    def load_cruises_from_pickle_file(self, file):
        logger.debug('Loading cruises at %s.' % file)
        self.cruises = util.io.object.load(file)
        logger.debug('Cruises loaded at %s.' % file)



import measurements.po4.woa.data13.regrid
import measurements.land_sea_mask.lsm
import util.io.np

def measurements_to_metos_boxes():
    from .constants import MEANS_FILE, NOBS_FILE, VARIANCES_FILE

    lsm = measurements.land_sea_mask.lsm.LandSeaMaskTMM(t_dim=12)
    z_values = measurements.po4.woa.data13.load.depth_values()
    (means, nobs, variances) = measurements.po4.woa.data13.regrid.measurements_to_land_sea_mask(lsm, z_values)

    for (value, file) in [[means, MEANS_FILE], [nobs, NOBS_FILE], [variances, VARIANCES_FILE]]:
        util.io.np.save(file, value, make_read_only=True, create_path_if_not_exists=True)
